import java.io.File;
import com.amazonaws.*;
import com.amazonaws.auth.PropertiesCredentials;
import com.amazonaws.services.s3.*;
import com.amazonaws.services.s3.model.*;

/** 
 * MPAWS
 * This class is used to interface with our Amazon S3 database
 * where all Magic Piano performances are stored. Performance MIDI
 * files are downloaded using this class after the resource keys are
 * obtained using MPSQL.
 * 
 * @author Nick Rudolfsky
 * @date Summer 2012
 * @version 1.0
 */
public class MPAWS 
{
	/** S3 singleton that represents a connection with our database */
	private static AmazonS3 s3 = null;
	
	/**
	 * Connect to the Amazon S3 database
	 * @return true is connected to DB successfully, false otherwise
	 */
	public static boolean connectToAWS()
	{
		try {
			s3 = new AmazonS3Client(new PropertiesCredentials(
					MPAWS.class.getResourceAsStream("AwsCredentials.properties")));
			
		} catch (Exception e) {
            printAWSException(e, "connectToAWS");
            return false;
		}
		
		return true;
	}
	
	/**
	 * Download a resource from AWS
	 * @param resource the resource to download
	 */
	public static boolean downloadResourceToFile(String resource, File localFile)
	{
		System.out.println("Downloading: " + resource);
		String bucket = "smule-resources";
		
		try {
			s3.getObject(new GetObjectRequest(bucket, resource), localFile);
			
		} catch (Exception e) {
            printAWSException(e, "downloadResource");
            return false;
		}
		
		return true;
	}
	
	
	/**
	 * Helper method to list all the AWS buckets in the DB
	 */
	public static void printAllBuckets()
	{
		System.out.println("Listing buckets");

		try {
			for (Bucket bucket : s3.listBuckets()) {
				System.out.println(" - " + bucket.getName());
			}
		} catch (Exception e) {
			printAWSException(e, "printAllBuckets");
		}
	}
	
	/**
	 * Helper method used to log an AWS exception
	 * @param e the exception to print
	 */
	public static void printAWSException(Exception e, String methodName)
	{
		System.out.println(methodName + " exception:");
		
		if (e instanceof AmazonServiceException)
		{
			AmazonServiceException ase = (AmazonServiceException)e;
			System.out.println("Error Message:    " + ase.getMessage());
            System.out.println("HTTP Status Code: " + ase.getStatusCode());
            System.out.println("AWS Error Code:   " + ase.getErrorCode());
            System.out.println("Error Type:       " + ase.getErrorType());
            System.out.println("Request ID:       " + ase.getRequestId());
		}
		else if (e instanceof AmazonClientException)
		{
			AmazonClientException ace = (AmazonClientException)e;
			System.out.println("Error Message: " + ace.getMessage());
		}
		else
		{
			System.out.println("Exception: " + e);
		}
	}
}
