import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;

/** 
 * MPDialogs
 * This class is used to display basic GUI dialogs such as
 * a performance downloader and progress dialog.
 * 
 * @author Nick Rudolfsky
 * @date Summer 2012
 * @version 1.0
 */
public class MPDialogs 
{
	/**
	 * Initialize all dialogs
	 */
	public static void initializeDialogs()
	{
		MPDownloadDialog.initialize();
		MPProgressDialog.initialize();
	}

	
	
	/**
	 * This class is used to download performances
	 */
	public static class MPDownloadDialog
	{
		/** Modal dialog */
		private static JDialog modalDialog;
		/** labels */
		private static JLabel songLabel;
		private static JLabel numberLabel;
		/** text fields */
		private static JTextField songField;
		private static JTextField numberField;
		/** buttons */
		private static JButton downloadButton;
		
		/**
		 * Initialize download dialog
		 */
		public static void initialize()
		{
			// create a dialog and panel to add GUI elements too
			modalDialog = new JDialog((JFrame)null, "Performance Downloader");
			modalDialog.setPreferredSize(new Dimension(500, 100));
			JPanel center_panel = new JPanel();
			
			// add the song label and text field
			songLabel = new JLabel("Song UID: ");
			center_panel.add(songLabel);
			songField = new JTextField("twinkle_twinkle_bundled");
			songField.setPreferredSize(new Dimension(400, 30));
			center_panel.add(songField);
			
			// add the number label and text field
			numberLabel = new JLabel("Amount to Download: ");
			center_panel.add(numberLabel);
			numberField = new JTextField("50");
			numberField.setPreferredSize(new Dimension(210, 30));
			center_panel.add(numberField);
			
			// add the download button
			downloadButton = new JButton("Download");
			downloadButton.addActionListener(new ActionListener() 
			{
				public void actionPerformed(ActionEvent arg0) {
					hitDownload();
				}
			});
			center_panel.add(downloadButton);

			// position the dialog
			modalDialog.setLocation(100, 100);
			modalDialog.toFront(); // raise above other java windows

			// add the panel to the dialog and pack
			modalDialog.getContentPane().add(center_panel, BorderLayout.CENTER);
			modalDialog.pack();
		}
		
		/**
		 * Called when download button is hit. do
		 */
		private static void hitDownload()
		{
			// get the information from text fields
			final String song_uid = songField.getText();
			final int performancesToDownload = Integer.parseInt(numberField.getText());
			
			// download performances on a background thread
			Thread background = new Thread(new Runnable() {
				@Override
				public void run() {
					// Find performance for this song in SQL database
					ArrayList<MPPerformance.MPMetaData> performances = MPSQL.getPerformancesForSong(song_uid, performancesToDownload);
					
					// alert user if no performnaces were found
					if (performances.size() <= 0)
						JOptionPane.showMessageDialog(null, "No Performances Found. Check song_uid.");
					
					// download each performance from Amazon S3
					MPDialogs.MPProgressDialog.showProgressDialog("Downloading Performances For\n" + song_uid, false);
					int perfIndex = 0;
					
					// iterate through each performance...
					for (MPPerformance.MPMetaData perfData : performances) {
						// update our progress dialog
						perfIndex++;
						MPDialogs.MPProgressDialog.updateProgress((float)perfIndex / (float)performances.size());
						
						// download performance to local file
						File localFile = new File(MPAnalyzer.GetPerfDirectory() + song_uid + "/" + perfData.resourceUUID);
						MPAWS.downloadResourceToFile(perfData.resourceUUID, localFile);
						
						// save meta data to file alongside MIDI file
						File metaFile = new File(MPAnalyzer.GetPerfDirectory() + song_uid + "/" + 
								perfData.resourceUUID.substring(0, perfData.resourceUUID.length() - 5) + ".meta");
						try {
							FileOutputStream fos = new FileOutputStream(metaFile);
							ObjectOutputStream oos = new ObjectOutputStream(fos);
							oos.writeObject(perfData);
							oos.close();
						} catch (FileNotFoundException e) {
							System.out.println("Warning: Couldn't write performance meta-data to file (no location info)");
						} catch (IOException e) {
							System.out.println("Warning: Couldn't write performance meta-data to file (no location info)");
						}
					}

					System.out.println("Total performances downloaded " + perfIndex);

					// dismiss progress dialog and refresh performance browser
					MPDialogs.MPProgressDialog.hideProgressDialog();
				}
			});
			background.start();
		}
		
		/**
		 * Display the download dialog. 
		 */
		public static void showDownloadDialog()
		{
			modalDialog.setVisible(true);
			modalDialog.pack();
		}
	}

	
	
	/**
	 * This is a progress dialog that can be called from background
	 * threads to indicate progress of work.
	 */
	public static class MPProgressDialog
	{
		/** Singleton modal dialog */
		private static JDialog modalDialog;
		/** Singleton label above progress bar */
		private static JLabel progressLabel;
		/** Singleton progress bar */
		private static JProgressBar progressBar;

		/**
		 * Initialize the GUI
		 */
		public static void initialize()
		{
			// create a dialog and panel to add GUI elements too
			modalDialog = new JDialog((JFrame)null, "Magic Piano Analyzer");
			modalDialog.setPreferredSize(new Dimension(600, 100));
			JPanel center_panel = new JPanel();

			// add a label
			progressLabel = new JLabel("Progress: ");
			center_panel.add(progressLabel);

			// add a progress bar
			progressBar = new JProgressBar(0, 100);
			progressBar.setPreferredSize(new Dimension(550, 40));
			progressBar.setStringPainted(true);
			progressBar.setValue(0); 
			center_panel.add(progressBar);

			// position the dialog
			modalDialog.setLocation(100, 300); // center on screen
			modalDialog.toFront(); // raise above other java windows

			// add the panel to the dialog and pack
			modalDialog.getContentPane().add(center_panel, BorderLayout.CENTER);
			modalDialog.pack();
		}

		/**
		 * Display the progress dialog. This will set progress to 0
		 * @param message message to display above dialog
		 * @param indeterminate should this bar be displayed as indeterminate?
		 */
		public static void showProgressDialog(String message, boolean indeterminate)
		{
			// set the label and clear the progress
			progressLabel.setText(message);
			progressBar.setValue(0);
			progressBar.setIndeterminate(indeterminate);

			modalDialog.setVisible(true);
			modalDialog.pack();
		}

		/**
		 * Display the progress dialog. This will set progress to 0
		 * @param message message to display above dialog
		 */
		public static void updateProgress(final float progress)
		{
			try {
				SwingUtilities.invokeAndWait(new Runnable() 
				{
					public void run() {
						int intProgress = (int)(100 * progress);
						progressBar.setValue(intProgress);
					}
				});
			} catch (InterruptedException e) {
				e.printStackTrace();
			} catch (InvocationTargetException e) {
				e.printStackTrace();
			}
		}

		/**
		 * Hide the progress dialog
		 */
		public static void hideProgressDialog()
		{
			modalDialog.setVisible(false);
		}
	}
}
