import java.io.Serializable;
import java.util.ArrayList;
import java.util.TreeMap;

/***
 * 
 * MPGeoCoord class.  Provide reverse geocoding for a set of coordinates.
 * 
 * @author jeffreysmith
 * 
 * v1.0 August 2012
 *
 */

public class MPGeoCoord2 implements Serializable, Comparable<MPGeoCoord2> {
	/** Version UID */
	public static final long serialVersionUID = 1;
	
	/*
	 * a hack of sorts to work around reverse geocode limits.  if we hit a limit, we start
	 * rounding all searches to this value, reducing dependency on future queries.  note it
	 * is important to not cache any rounded results.
	 */
	static boolean geo_approx_coords = false;
	
	/*
	 * rounding value for approximate hack.  if you change this value, you need to flush
	 * and re-sort the geocache.
	 */
	private static final float geo_scale_approximate = MPAnalyzer.GetConfig().GEO_SCALE_APPROX;
	// private static float geo_scale_max = (float)1000000.0;
	private static final float geo_scale = geo_scale_approximate;
	static boolean trace_geo_compare = false;

	float n_lat;
	float n_long;
	
	String code;
	String address;
	String countryNameCode;
	String countryName;
	String administrativeAreaName;
	String localityName;
	String thoroughfareName;
	String postalCodeNumber;
	
	// we use these fields to cache whether the local is a city
	// given google returns various names of a city, we will also store
	// a common city name, which is retrieved from the city list (in MPGeocoder class)
	boolean is_city;
	String commonLocalityName;
	
	static void TraceGeoCompare(boolean value)
	{
		trace_geo_compare = value;
	}
	
	protected MPGeoCoord2(MPGeoCoord2 coord)
	{
		n_lat = coord.n_lat;
		n_long = coord.n_long;
		code = coord.code;
		address = coord.address;
		countryNameCode = coord.countryNameCode;
		countryName = coord.countryName;
		administrativeAreaName = coord.administrativeAreaName;
		localityName = coord.localityName;
		thoroughfareName = coord.thoroughfareName;
		postalCodeNumber = coord.postalCodeNumber;
		is_city = coord.is_city;
		commonLocalityName = coord.commonLocalityName;
	}
	
	protected MPGeoCoord2(String latitude, String longitude)
	{
		n_lat = Float.parseFloat(latitude);
		n_long = Float.parseFloat(longitude);
	}
	
	private static int geoCoordCompareTo(float x1, float x2, float y1, float y2) 
	{
		if (geo_approx_coords) {
			x1 = Math.round(x1 * geo_scale);
			x2 = Math.round(x2 * geo_scale);
			y1 = Math.round(y1 * geo_scale);
			y2 = Math.round(y2 * geo_scale);
		}

		int result;
		if (x1 < x2) {
			result = -1;
		} else if (x1 > x2) {
			result = 1;
		} else if (y1 < y2) {
			result = -1;
		} else if (y1 > y2) {
			result = 1;
		} else {
			result = 0;
		}
		
		if (trace_geo_compare) {
			System.out.println("geocompare: " + result + " " + y1 + "," + x1 + " " + y2 + "," + x2);
		}
		return result;
	}
	
	public int compareTo(MPGeoCoord2 inst)
	{
		return geoCoordCompareTo(n_long, inst.n_long, n_lat, inst.n_lat);
	}
	
	public boolean isGeoCodeNull()
	{
		return n_lat == 0 && n_long == 0;
	}
	
	public String GetGeoString()
	{
		String s = n_lat + ", " + n_long + " ";
		if (countryNameCode != null) {
			s += countryNameCode + ", ";
		}
		if (administrativeAreaName != null) {
			s += administrativeAreaName + ", ";
		}
		if (localityName != null) {
			s += localityName + ", ";
		}
		if (address != null) {
			s += address;
		}
		return s;
	}
	
	public MPJsonWriter KeyPairs()
	{
		MPJsonWriter obj = new MPJsonWriter();
		obj.JsonPut("code", code);
		obj.JsonPut("address", address);
		obj.JsonPut("countryNameCode", countryNameCode);
		obj.JsonPut("countryName", countryName);
		obj.JsonPut("administrativeAreaName", administrativeAreaName);
		obj.JsonPut("localityName", localityName);
		obj.JsonPut("thoroughfareName", thoroughfareName);
		obj.JsonPut("postalCodeNumber", postalCodeNumber);
		obj.JsonPut("commonLocalityName", commonLocalityName);
		obj.JsonPut("is_city", is_city);
		return obj;
	}
	
	static final String[] redStates = {"AL", "AK", "AZ", "AR", "CO",  
					"GA", "ID", "IN", "KS", "KY", "LA", "MS", 
					"MO", "MT", "NE", "NV", "NH", "NC", "ND", 
					"OH", "OK", "PR", "SC", "SD", "TN", "TX", 
					"UT", "VA", "WV", "WY"};
	
	static final String[] blueStates = {"CA", "CT", "DE", "HI", "FL",
					"IL", "IA", "ME", "MD", "MA", "MI", "MN", 
					"NJ", "NM", "NY", "OR", "PA", "RI", "VT", 
					"WA", "WI" };
		
	static final String[] swingStates = {"CO", "NV", "IA", "WI", "VA", "OH", 
		"FL", "NC", "NH"};

	static final String[] usStates = {"CA", "CT", "DE", "FL", "HI", 
		"IL", "IA", "ME", "MD", "MA", "MI", "MN", 
		"NJ", "NM", "NY", "OR", "PA", "RI", "VT", 
		"WA", "WI",
		"AL", "AK", "AZ", "AR", "CO", 
		"GA", "ID", "IN", "KS", "KY", "LA", "MS", 
		"MO", "MT", "NE", "NV", "NH", "NC", "ND", 
		"OH", "OK", "PR", "SC", "SD", "TN", "TX", 
		"UT", "VA", "WV", "WY" };
	
	static public String[] getRedStates()
	{
		return redStates;
	}

	static public String[] getBlueStates()
	{
		return blueStates;
	}
	
	static public String[] getSwingStates()
	{
		return swingStates;
	}
	
	static public String[] getStates()
	{
		return usStates;
	}
	
	public boolean isUSState(String s) 
	{
		return isUS() && administrativeAreaName != null && administrativeAreaName.equalsIgnoreCase(s);
	}

	public boolean isRedState() 
	{
		if (!isUS()) {
			return false;
		}
		
		for (String s : redStates) {
			if (s.equalsIgnoreCase(administrativeAreaName)) {
				return true;
			}
		}
	
		return false;
	}
	
	public boolean isBlueState()
	{
		return isUS() && !isRedState();
	}
	
	public boolean isSwingState()
	{
		if (!isUS() || administrativeAreaName == null) {
			return false;
		}
		
		for (String s : swingStates) {
			if (s.equalsIgnoreCase(administrativeAreaName)) {
				return true;
			}
		}
	
		return false;
	}
	
	public boolean isCountry(String s)
	{
		if (isGeoCodeNull()) {
			return false;
		}
		
		return countryNameCode != null && countryNameCode.equalsIgnoreCase(s);		
	}
	
	public String getCountry()
	{
		return countryNameCode;
	}

	static String[][] country_pairs = {
		{"CN", "China"},
		{"RU", "Russia"},
		{"TH", "Thailand"}, 
		{"KR", "South Korea"},
		{"JP", "Japan"},
		{"TW", "Taiwan"},
		{"SG", "Singapore"},
		{"IL", "Israel"},
		{"EG", "Egypt"},
		{"SA", "Saudi Arabia"},
		{"VN", "Vietnam"},
		{"BH", "Bahrain"},
		{"IQ", "Iraq"},
		{"TR", "Turkey"},
		{"BN", "Brunei"},
		{"LV", "Latvia"},
		{"SK", "Slovakia"},
		{"MN", "Mongolia"},
		{"CY", "Cyprus"},
		{"PL", "Poland"},
		{"CZ", "Czech Republic"},
		{null, null}
	};

	static final String[][] state_pairs = {
		{"AL", "Alabama"},
		{"AK", "Alaska"},
		{"AZ", "Arizona"},
		{"AR", "Arkansas"},
		{"CA", "California"},
		{"CO", "Colorado"},
		{"CT", "Connecticut"},
		{"DE", "Delaware"}, 
		{"DC", "District of Columbia"}, 
		{"FL", "Florida"},
		{"GA", "Georgia"},
		{"HI", "Hawaii"},
		{"ID", "Idaho"},
		{"IL", "Illinois"},
		{"IN", "Indiana"},
		{"IA", "Iowa"},
		{"KS", "Kansas"},
		{"KY", "Kentucky"},
		{"LA", "Louisiana"},
		{"ME", "Maine"},
		{"MD", "Maryland"},
		{"MA", "Massachusetts"},
		{"MI", "Michigan"},
		{"MN", "Minnesota"},
		{"MS", "Mississippi"},
		{"MO", "Missouri"},
		{"MT", "Montana"},
		{"NE", "Nebraska"},
		{"NV", "Nevada"},
		{"NH", "New Hampshire"},
		{"NJ", "New Jersey"},
		{"NM", "New Mexico"},
		{"NY", "New York"},
		{"NC", "North Carolina"},
		{"ND", "North Dakota"},
		{"OH", "Ohio"},
		{"OK", "Oklahoma"},
		{"OR", "Oregon"},
		{"PA", "Pennsylvania"},
		{"RI", "Rhode Island"},
		{"SC", "South Carolina"},
		{"SD", "South Dakota"},
		{"TN", "Tennessee"},
		{"TX", "Texas"},
		{"UT", "Utah"},
		{"VT", "Vermont"},
		{"VA", "Virginia"},
		{"WA", "Washington"},
		{"WV", "West Virginia"},
		{"WI", "Wisconsin"},
		{"WY", "Wyoming"},
		{null, null}
	};
	
	static TreeMap<String, String> state_map = null;
	
	private static void initStateMap()
	{
		state_map = new TreeMap<String,String>();
		for (int i = 0; state_pairs[i][0] != null; i++) {
			state_map.put(state_pairs[i][0], state_pairs[i][1]);
		}
		
	}
	
	public static String MapStateName(String state_abbrev) 
	{
		if (state_map == null) {
			initStateMap();
		}
		
		if (state_abbrev == null || state_abbrev.isEmpty()) {
			return state_abbrev;
		}
		
		String s = state_map.get(state_abbrev.toUpperCase());
		if (s != null && !s.isEmpty()) {
			return s;
		}
		return state_abbrev;
	}
		
	static TreeMap<String, String> country_map = null;

	private static void initCountryMap()
	{
		country_map = new TreeMap<String,String>();
		for (int i = 0; country_pairs[i][0] != null; i++) {
			country_map.put(country_pairs[i][0], country_pairs[i][1]);
		}
		
	}

	private String fullNamefromCountryCode()
	{
		if (country_map == null) {
			initCountryMap();
		}
		String s = country_map.get(countryNameCode);
		if (s != null && !s.isEmpty()) {
			return s;
		}
		
		return countryNameCode;
	}
	
	private Boolean isAscii(String name)
	{
		if (name == null) {
			return false;
		}
		for (int i = 0; i < name.length() - 1; i++) {
			int c = name.charAt(i);
			if (c != ' ') {
				if (c < 'A' || c > 'z') {
					return false;
				} else if (c > 'Z' && c < 'a') {
					return false;
				}
			}
		}
		return true;
	}

	public String getFullCountryName()
	{
		if (!isAscii(countryName)) { 
			return fullNamefromCountryCode();
		}
		
		return countryName; // + " (" + countryNameCode + ")";
	}
	
	public String getState()
	{
		return administrativeAreaName;
	}
	
	public String getCity()
	{
		return localityName;
	}
	
	public String getCommonCityName()
	{
		return commonLocalityName;
	}
	
	public boolean isState()
	{
		return isUS() && administrativeAreaName != null;
	}
	
	public boolean isState(String thestate)
	{
		return administrativeAreaName != null && administrativeAreaName.equalsIgnoreCase(thestate);
	}
	
	public boolean isChina()
	{
		return isCountry("CN");
	}
	
	public boolean isUS()
	{
		return isCountry("US");
	}
	
	public boolean isCanada()
	{
		return isCountry("CA");
	}
	
	public boolean isMX()
	{
		return isCountry("MX");
	}
	
	public boolean isCity()
	{
		return is_city; // MPGeocoder.IsCity(this);
	}
	
	public boolean Asia()
	{
		return (n_lat >= -13 && (n_long > 40 && n_long <= 180));
	}

	public boolean Europe()
	{
		return (n_lat >= 30 && (n_long >= -30 && n_long <= 40));
	}
	
	public boolean NorthAmerica()
	{
		return ((n_long >= -180 && n_long <= -30) &&
				((n_lat >= 32.5) ||
				 (n_lat >= 25 && !SouthAmerica()) ||
				 (n_lat >= 16 && (n_long >= -180 && n_long < -118)))); 
		
		/*
		return (n_lat >= 15 && (n_long >= -180 && n_long <= -92)) ||
				(n_lat >= 18 && (n_long >= -180 && n_long < -30));
		*/
		
	}
	
	public boolean SouthAmerica()
	{
		// adding mexico
		return (n_lat < 25 && (n_long > -118 && n_long < -20)) ||
				
				/* this is the texas border */
				(n_lat < 26 && (n_long > -118 && n_long <= -97)) ||
				(n_lat < 27 && (n_long > -118 && n_long <= -99)) ||
				(n_lat < 28 && (n_long > -118 && n_long <= -100)) ||
				(n_lat < 29 && (n_long > -118 && n_long <= -101)) ||
				(n_lat < 30 && (n_long > -118 && n_long <= -105)) ||
				(n_lat < 32 && (n_long > -118 && n_long <= -106)) ||
				(n_lat < 32.5 && (n_long > -118 && n_long <= -114.8));
				
		
//		return (n_lat < 25 && (n_long > -92 && n_long <= -20));
	}
	
	public boolean Africa()
	{
		return (n_lat < 30 && (n_long > -20 && n_long < 60));
	}
	
	public boolean Australia()
	{
		return (n_lat < -13 && 
				((n_long > 110 && n_long < 180) || 
				 (n_long > -180 && n_long < 100)));
	}
	 	
	static public String geoIndexToString(int i)
	{
		String s = "";
		switch (i) {
		case 0:
			s = "Asia";
			break;
		case 1:
			s = "Europe";
			break;
		case 2:
			s = "North America";
			break;
		case 3:
			s = "South America";
			break;
		case 4:
			s = "Africa";
			break;
		case 5:
			s = "Australia";
			break;
		default:
			break;
		}
		return s;
	}
	
	public short geoIndex()
	{		
		short i = 0;
		if (!isGeoCodeNull()) {
			if (Asia()) {
				i = 0;
			} else if (Europe()) {
				i = 1;
			} else if (NorthAmerica()) {
				i = 2;
			} else if (SouthAmerica()) {
				i = 3;
			} else if (Africa()) {
				i = 4;
			} else if (Australia()) {
				i = 5;
			}
		}
		return i;
	}
	
	public String getGeoString()
	{
		return geoIndexToString(geoIndex());
	}
	
}
