import java.io.*;

import javax.sound.midi.*;

/** 
 * MPMIDI
 * This class is used to parse performances which are
 * stored in MIDI format. It parses the performances manually
 * using java's built-in MIDI APIs. MIDI files are parsed into
 * MPPerformance objects which can then be analyzed.
 * 
 * @author Nick Rudolfsky
 * @date Summer 2012
 * 
 * update jeffreysmith august 2012
 * 
 * @version 1.0
 */
public class MPMIDI {	
	/** Private variables used to parse MIDI file */
	private static final int NOTE_ON = 0x90;
	private static final int NOTE_OFF = 0x80;
	private static final boolean DEBUG_MIDI_PARSER = false;
	private static int DEBUG_MIDI_LEVEL = 1;

	/** The currently playing sequencer */
	private static Sequencer currentSequencer = null;
	
	static private void DebugParser(String s)
	{
		if (DEBUG_MIDI_PARSER) {
			System.out.println(s);
		}
	}
	
	static private void DebugParser(int level, String s)
	{
		if (level >= DEBUG_MIDI_LEVEL) {
			DebugParser(s);
		}
	}
	
	static private MPNoteEvent fireNoteOn(ShortMessage sm, long ms)
	{
		// parse the information
		int key = sm.getData1();		// MIDI value
		int velocity = sm.getData2();	// Velocity
		MPNoteEvent noteEvent = new MPNoteEvent(key, velocity, ms);
		
		DebugParser("note on " + noteEvent.getMIDIValue() + " " + noteEvent.getPitch() + " " + ms);
		return noteEvent;
	}
	
	static private void fireNoteOff(ShortMessage sm, float ms)
	{
		// parse the information
		int key = sm.getData1();
		DebugParser(2, "Note off, " + " key=" + key);
	}
	
	static private int fireTempoEvent(MetaMessage meta)
	{
		// calculate the BPM
		byte[] data = meta.getData();
    	long mask = 0x00FF;
    	int b0 = (int)(mask & data[0]);
    	int b1 = (int)(mask & data[1]);
    	int b2 = (int)(mask & data[2]);
		long tempo = (b0 << 16) + (b1 << 8) + b2;
		long microSecondsPerMinute = (long)60000000;
		long BPM = (microSecondsPerMinute / tempo);
		DebugParser("Tempo: " + BPM);

		return (int)BPM;
	}
	
	private static long ticksToMS(long ticks, long BPM)
	{
		return ticks * (long)60 / BPM;
	}

	public static void fireEvent(MPMidiData newPerf, float bpmRate, MidiEvent event)
	{
		MidiMessage message = event.getMessage();
		
		// get the timing info for this event
		DebugParser(3, "@" + event.getTick() + " ");

		// parse note message
		if (message instanceof ShortMessage) {
			ShortMessage sm = (ShortMessage) message;
			
			DebugParser(3, "Channel: " + sm.getChannel() + " ");

			// if this is a NOTE_ON event...
			if (sm.getCommand() == NOTE_ON) {
				MPNoteEvent noteEvent = fireNoteOn(sm, ticksToMS(event.getTick(), newPerf.BPM));
				newPerf.noteEvents.add(noteEvent);

				// or a NOTE_OFF event...
			} else if (sm.getCommand() == NOTE_OFF) {
				fireNoteOff(sm, ticksToMS(event.getTick(), newPerf.BPM));

				//TODO: seek through existing note events and set a duration for the last NOTE_ON of this NOTE_OFF
				// only other ShortMessage is controller event...
			} else {
				
				DebugParser(2, "Controller:" + sm.getData2());
			}

		// This is some kind of other meta data such as tempo
		} else {
			try {
				MetaMessage meta = (MetaMessage)message;

				// Got a tempo message
				if (meta.getType() == 81) {
					newPerf.BPM = Math.round(bpmRate * fireTempoEvent(meta));
	
				// got a time signature message
				} else if (meta.getType() == 88) {
					DebugParser("Time Signature: " + meta.getData());
					
				} else {
					DebugParser(3, "Unknown meta message " + meta.getData());
				
				}
			} catch (Exception e) {
				
				System.err.println(e);
			}
		}
	}
	
	private static Sequence loadSequence(DataInputStream dis)
	{
		Sequence s = null;
		try {
			s = MidiSystem.getSequence(dis);
		} catch (InvalidMidiDataException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}		
		return s;
	}

	static short parseMidiStream(MPMidiData newPerf, float bpmRate, DataInputStream dis)
	{
		// initialize the MIDI sequence from the MIDI file
		Sequence sequence = loadSequence(dis);
		if (sequence == null) {
			return -1;
		}
		
		// iterate through all the tracks in the MIDI file
		int trackNumber = 0;
		for (Track track :  sequence.getTracks()) {
			trackNumber = trackNumber + 1;
			DebugParser("Track " + trackNumber + ": size = " + track.size());

			// iterate through all events in the track
			for (int i = 0; i < track.size(); i++) { 
				// get the message associate with this event
				MidiEvent event = track.get(i);
				fireEvent(newPerf, bpmRate, event);
			}
		}
		
		return 0;
	}
	
	/**
	 * This method parses a MIDI file into an MPPerformance object.
	 * @param midiFile
	 * @return A rendered -1 if error, zero otherwise
	 */
	public static short parseMIDI(MPMidiData newPerf, float bpmRate, File midiFile) 
	{

		FileInputStream fis = null;
		BufferedInputStream bis = null;
		DataInputStream dis = null;		
		short result = -1;
		
		try {
			fis = new FileInputStream(midiFile);
			bis = new BufferedInputStream(fis);
			dis = new DataInputStream(bis);
			result = parseMidiStream(newPerf, bpmRate, dis);
			
		} catch (IOException e) {
			e.printStackTrace();
			return -1;
		} finally {
			try {
				fis.close();
				bis.close();
				dis.close();
			} catch (IOException ex) {
				ex.printStackTrace();
			}
		}
		
		// nuke the midi if the parse failed -- assume bad midi
		if (result != 0) {
			midiFile.delete();
			File metaFile = new File(midiFile.getPath().substring(0, midiFile.getPath().length() - 5) + ".meta");
			metaFile.delete();
		}
		
		return result;
	}

	/**
	 * Method that can play a MIDI file on a background thread
	 * @param midiFile the MIDI file to play
	 */
	public static void playMIDI(final File midiFile)
	{	
		// safety check to turn off currently playing performance
		stopMIDI();
		
		// play the MIDI file on background thread
		Thread backgroundThread = new Thread(new Runnable() {
			public void run() {
				try {
					// start MIDI sequence from file
					currentSequencer = MidiSystem.getSequencer();
					currentSequencer.open();
					currentSequencer.setSequence(MidiSystem.getSequence(midiFile));
					currentSequencer.start();
				} catch (IOException e) {
					e.printStackTrace();
				} catch (MidiUnavailableException e) {
					e.printStackTrace();
				} catch (InvalidMidiDataException e) {
					e.printStackTrace();
				}
			}
		});
		backgroundThread.start();
	}

	/**
	 * Stop the currently playing MIDI file
	 */
	public static void stopMIDI()
	{
		// stop MIDI sequencer if it's running
		if (currentSequencer != null &&  currentSequencer.isOpen()) {
			currentSequencer.stop();
			currentSequencer.close();
		}
	}

	/**
	 * Helper method used to test our parsing
	 * @param midiFile the MIDI file to parse
	 * @return
	 * 
	public static void testParsing(File midiFile)
	{
		// parse the MIDI using our custom parser
		MPPerformance performance = parseMIDI(midiFile);

		// print and play the music string obtained from our parser
		String perfString = performance.getMusicString();
		System.out.println(perfString);
		Player player = new Player();
		player.play(perfString);

		// print out the JFugue Music String for the build in parser (not being used)
		/*try {
			Sequence midiSequence = MidiSystem.getSequence(midiFile);
			MidiParser midiParser = new MidiParser();
			MusicStringRenderer renderer = new MusicStringRenderer();
			midiParser.addParserListener(renderer);
			midiParser.parse(midiSequence);
			String musicString = renderer.getPattern().getMusicString();
			System.out.println(musicString);
		} catch (InvalidMidiDataException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
*/

}
