import java.io.Serializable;



/** 
 * MPNoteEvent
 * This class represents a single note hit during a performance. It contains
 * information about the note such as the pitch and the time at which the note
 * was hit. This information is represented using JFugue Note and Time objects,
 * so the JFugue Music Programming API can be used if desired.
 * 
 * @author Nick Rudolfsky
 * 
 * @Edited jeffreysmith:  bug fix; ticks != MS.  ticks * 60/BPM = MS.
 * @date Summer 2012
 * @version 1.0
 * 
 * @Edited jeffreysmith:  rewrote and eliminated all references to jfugue -- that jar is just too buggy
 * @date October 2012
 * @version 1.1
 * 
 */
public class MPNoteEvent implements Serializable {
	/** Version UID */
	public static final long serialVersionUID = 1;

	private static final String[] NOTE_NAMES = {"C", "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "B"};

	protected int key;
	protected int velocity;
	protected long timeinms;
	
	/**
	 * Constructor
	 * @param note The JFugue note object
	 * @param time The JFugue time object
	 */
	public MPNoteEvent(int notekey, int notevelocity, long ms)
	{
		key = notekey;
		velocity = notevelocity;
		timeinms = ms;
	}
		
	static String getPitch(int key) 
	{
		int n = key % 12;				// Piano Note
		String noteName = NOTE_NAMES[n];

		return noteName;
	}
	/**
	 * Get the pitch for this not event
	 * @return the pitch of this note event
	 */
	public String getPitch()
	{
		return getPitch(key);
	}
	
	public int getVelocity()
	{
		return velocity;
	}
	
	public int getOctave()
	{
		int octave = (key / 12)-1;		// Octave
		return octave;
	}
	
	/**
	 * Get the MIDI value for this note event
	 * @return the MIDI value of this note event
	 */
	public int getMIDIValue()
	{
		return key;
	}
	
	/**
	 * Get the time of this note event in milliseconds
	 * @return the time of this note event in milliseconds
	 */
	public long getTimeMS()
	{
		return timeinms;
	}
	
	/**
	 * Get the time of this note event in seconds
	 * @return the time of this note event in seconds
	 */
	public long getTimeS()
	{
		return (getTimeMS() / 1000);
	}
	
	public int compareTo(MPNoteEvent anotherInstance) 
	{
		long result = getMIDIValue() - anotherInstance.getMIDIValue();
		if (result == 0) {
			result = getVelocity() - anotherInstance.getVelocity();
			if (result == 0) {
				result = Math.round(getTimeMS() - anotherInstance.getTimeMS());
			}
		}
		return (int)result;
	}
}
