import javax.swing.*;
import javax.swing.event.*;
import javax.swing.tree.*;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.io.File;

/** 
 * MPPerfBrowser
 * This is the main GUI of the application. It contains a file browser
 * to navigate the downloaded performances (located in the "perfs" directory
 * of this eclipse project folder. It also provides a menu to bring up the performance 
 * downloader and buttons to perform actions on the selected performance.
 * 
 * @author Nick Rudolfsky & jeffreysmith
 * @date Summer 2012
 * @version 1.0
 * 
 * version 1.1 nov '12:  pruned the directory walking stuff.  too expensive with lots of files.
 * 
 */
public class MPPerfBrowser 
{
	/** The main frame of the window */
	private static JFrame mainFrame;
	/** The file browser */
	private static MPFileTree fileTree;
	
	/**
	 * Initialize the GUI
	 */
	public static void initializeGUI()
	{
		// create the main frame
		mainFrame = new JFrame("Magic Piano Performance Analyzer");
		mainFrame.setLayout(new FlowLayout());
		mainFrame.setLocation(700, 100);
		
		// add the performance browser
		fileTree = new MPFileTree();
		fileTree.MPFileInit();
		fileTree.setPreferredSize(new Dimension(500, 500));
		mainFrame.getContentPane().add(fileTree);

		// create the download menu
		JMenu downloadMenu = new JMenu("Download");
		JMenuItem downloadItem = new JMenuItem("Download Performances");
		downloadItem.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				MPDialogs.MPDownloadDialog.showDownloadDialog();
			}
		});
		downloadMenu.add(downloadItem);

		// add the menu to the frame
		JMenuBar menuBar = new JMenuBar();
		menuBar.add(downloadMenu);
		mainFrame.setJMenuBar(menuBar);

		// close the program when we close this main window
		mainFrame.addWindowListener(new WindowListener() {
			public void windowOpened(WindowEvent e) {}
			public void windowIconified(WindowEvent e) {}
			public void windowDeiconified(WindowEvent e) {}
			public void windowDeactivated(WindowEvent e) {}
			public void windowActivated(WindowEvent e) {}
			public void windowClosed(WindowEvent e) {}
			public void windowClosing(WindowEvent e) { System.exit(1); }
		});

		// create a panel for buttons
		JPanel buttonPanel = new JPanel();
		buttonPanel.setPreferredSize(new Dimension(150, 500));

		// add an analyze button to analyze performances
		JButton analyzeButton = new JButton("Analyze");
		analyzeButton.setPreferredSize(new Dimension(120, 30));
		analyzeButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) 
			{
				// verify we selected a file and that it's not a directory
				if ( fileTree.currentFile != null && !fileTree.currentFile.isDirectory() ) {
					// parse the selected file and analyze it
					MPPerformance perf = new MPPerformance(null, fileTree.currentFile);
					if (perf.LoadPerformance()) {
						perf.printAllNotes();
					}
					
				} else if ( fileTree.currentFile != null && fileTree.currentFile.isDirectory() ) {

					// analyze all songs in directory
					boolean write_output = true;
					MPSong songs = new MPSong(fileTree.currentFile, MPAnalyzer.GetPerfMax(), write_output);
					songs.LoadSong();
					songs.AnalyzeSongData();
					songs.Close();
				}
			}
		});
		buttonPanel.add(analyzeButton);

		// add a play button to play performances
		JButton playButton = new JButton("Play");
		playButton.setPreferredSize(new Dimension(120, 30));
		playButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) 
			{
				// verify we selected a file and that it's not a directory
				if ( fileTree.currentFile != null && !fileTree.currentFile.isDirectory() )
				{
					// play the current file
					MPMIDI.playMIDI(fileTree.currentFile);
				}
			}
		});
		buttonPanel.add(playButton);

		// add a play button to play performances
		JButton stopButton = new JButton("Stop");
		stopButton.setPreferredSize(new Dimension(120, 30));
		stopButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) 
			{
				MPMIDI.stopMIDI();
			}
		});
		buttonPanel.add(stopButton);

		// add the buttons to the main frame
		mainFrame.getContentPane().add(buttonPanel);

		// display the frame
		mainFrame.setVisible(true);
		mainFrame.pack();
	}
	

	/**
	 * A subclasses JFileTree for displaying downloaded performances
	 */
	private static class MPFileTree extends JPanel 
	{
		/** The currently selected File */
		public File currentFile = null;
		
		/** the underlying JTree */
		private JTree fileTree;

		/** suppress serializable warning by implementing versionUID */
		private static final long serialVersionUID = 1;

		/** 
		 * Constructor for our custom tree
		 */
		public MPFileTree() 
		{
		}
		
		public void MPFileInit()
		{
			setLayout(new BorderLayout());

			System.out.println("init directory of files...");
			// Make a tree with top level director
			File perfDirectory = new File(MPAnalyzer.GetPerfDirectory());
			fileTree = new JTree(addDirectory(perfDirectory));
			
			System.out.println("init directory done");
			
			// Add a listener for file selection
			fileTree.addTreeSelectionListener(new TreeSelectionListener() 
			{
				public void valueChanged(TreeSelectionEvent e) 
				{
					// build path to the current file
					String path = "../perfs/";
					for (int i = 0; i < e.getPath().getPathCount(); i++) 
					{
						path += e.getPath().getPathComponent(i);
						if (i != e.getPath().getPathCount() - 1)
							path += "/";
					}

					// set the current file
					currentFile = new File(path);
				}
			});

			// Lastly, put the JTree into a JScrollPane.
			JScrollPane scrollpane = new JScrollPane();
			scrollpane.getViewport().add(fileTree);
			add(BorderLayout.CENTER, scrollpane);
		}

		/** 
		 * Add nodes to the file browser recursively 
		 * @param curTop the parent node or null if this is root node
		 * @param dir the directory we are adding
		 */
		private DefaultMutableTreeNode addDirectory(File dir) 
		{
			DefaultMutableTreeNode dirRoot = new DefaultMutableTreeNode(dir.getName());
			
			// create this directories node
			String[] directoryItems = dir.list();

			// Iterate over all directory items, recursively adding directories and storing files
			for (int i = 0; i < directoryItems.length; i++) {
				String directoryItem = directoryItems[i];

				// determine path for new item
				String newPath = dir.getPath() + File.separator + directoryItem;

				// determine if directory or file
				File fileItem = new File(newPath);
				if ( fileItem.isDirectory() ) {
					DefaultMutableTreeNode curDir = new DefaultMutableTreeNode(fileItem.getName());
					dirRoot.add(curDir);
				}
			}

			return dirRoot;
		}
	}
}
