import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.Serializable;
import java.util.*;

/** 
 * MPPerformance
 * This class represents a full Magic Piano Performance which
 * has been parsed from the performance MIDI file. It contains
 * collections of MPNoteEvents which chronologically represent
 * the notes hit during a performance.
 * 
 * @author Nick Rudolfsky
 * @date Summer 2012
 * @version 1.0
 * 
 * @author jeffreysmith
 * @date August 2012  
 * @version 1.1 
 */

public class MPPerformance extends MPOriginalPerformance implements Comparable<MPPerformance> {
	/** The meta data (location data) of this performance */
	public MPMetaData metaData;
	
	protected MPGeoCoord2 perfLoc;
	protected MPPerfDate perfDate;
	
	// used to cache performance clusters
	protected int unique_id_2;
	
	protected void PermanentlyDeleteFile()
	{
		File tmpfile = new File(performanceFile.getPath().substring(0, performanceFile.getPath().length() - 5) + ".midi");
		tmpfile.delete();
		tmpfile = new File(performanceFile.getPath().substring(0, performanceFile.getPath().length() - 5) + ".cache");
		tmpfile.delete();
		tmpfile = new File(performanceFile.getPath().substring(0, performanceFile.getPath().length() - 5) + ".meta");
		tmpfile.delete();
		Clear();
	}
	
	protected void setUniqueID2(int i)
	{
		unique_id_2 = i;
	}
	
	protected int getUniqueID2()
	{
		return unique_id_2;
	}
		
	protected void initMeta()
	{
		if ( loadMetaData() != null ) {
			
			perfLoc = MPGeocoder.reverseGeocode(metaData.latitude, metaData.longitude);
			// nuke non-geocode data
			if (perfLoc.isGeoCodeNull()) {
				System.out.println("no geo data:  deleting " + performanceFile.getPath());
				PermanentlyDeleteFile();				
			}
			perfDate = new MPPerfDate(metaData.perfDate);
			
		} else {
			System.out.println("Failed to load meta data:  deleting " + performanceFile.getPath());
			PermanentlyDeleteFile();
		}
	}

	public MPPerformance(MPSongIO sIO, File midiFile)
	{
		super(sIO, midiFile);
		initMeta();		
	}
	
	public int getUserID()
	{
		return metaData.uniqueUserID;
	}
	
	public MPGeoCoord2 getLoc()
	{
		return perfLoc;
	}
	
	// output the meta data of the performance in json format
	public void WriteMetaData()
	{
		MPJsonWriter writer = new MPJsonWriter();
		
		// output time/date
		writer.JsonPut("date", perfDate.getString());
		
		// output GPS
		String GPS = perfLoc.n_lat + "," + perfLoc.n_long;
		writer.JsonPut("GPS", GPS);
				
		// output reverse geocode data
		MPJsonWriter reverse_GPS = perfLoc.KeyPairs();
		writer.JsonPut("Location", reverse_GPS);

		// output user identifier
		writer.JsonPut("user-id", getUserID());
		
		String jsonpath = performanceFile.getPath().substring(0, performanceFile.getPath().length() - 5) + ".json";
		writer.JsonWrite(jsonpath);
	}
	
	static public String OutputFormat()
	{
		return "UID\tdate\tslope\tstddev\tpolarity\tinflection\tregion\tGPS\tmidi\t\t";
	}
	
	public String getString()
	{
		/*
		String s = getUserID() + "\t" + perfDate.getString() + "\t" + getSlope() + "\t" + sortNoteValue;
		s = s + "\t" + perfPolarity + "\t" + perfInflection;
		s = s + "\t" + perfLoc.countryNameCode + " " + perfLoc.administrativeAreaName;
		s = s + "\t" + getLoc().n_lat + " " + getLoc().n_long;
		s = s + "\t" + performanceFile.getName() + "\t";
		*/
		String s = getUserID() + "\t" + perfDate.getString() + "\t" + getSlope();
		s = s + "\t" + perfLoc.countryNameCode;
		s = s + "\t" + performanceFile.getName() + "\t";
		return s;
	}
	
	public void Report(MPOutputStream out)
	{
		// out.MPWrite(getString() + "\t");
		out.MPWrite(performanceFile.getName() + "\t" + sortNoteValue + "\t" + getLoc().n_lat + "," + getLoc().n_long + "\n");
		for (int i = 0; i < getNumNotes(); i++) {
			out.MPWrite(noteDur.get(i) + "\t");
			// out.MPWrite(noteSlope.get(i) + "\t");
		}
		out.MPWriteLn("");
		for (int i = 0; i < getNumNotes(); i++) {
			// out.MPWrite(midi.noteEvents.get(i).getPitch() + "\t");
			out.MPWrite(midi.noteEvents.get(i).getMIDIValue() + "\t");
		}
		out.MPWriteLn("");
	}
	
    public int compareTo(MPPerformance anotherInstance) 
    {
    	int result = getUserID() - anotherInstance.getUserID();
    	if (result == 0) {
    		result = perfDate.compareTo(anotherInstance.perfDate);
    		if (result == 0) {
    			result = perfLoc.compareTo(anotherInstance.perfLoc);
    			if (result == 0) {
    				result = SongLength() - anotherInstance.SongLength();
    				for (int i = 0; result == 0 && i < SongLength() ; i++) {
    					result = midi.noteEvents.get(i).compareTo(anotherInstance.midi.noteEvents.get(i));
    				}
    			}
    		}
    	}
        return result;
    }

    static public String SlopeFormat()
	{
		return "UID\t" + "date\t" + "slope\t" + "stddev\t";

	}
    
	static final double[] truslit_open = {488, 536.8, 488, 488, 414.8, 439.2, 879.3, 414.8, 390.4, 439.2, 488, 536.8, 585.6, 1172.4, 0.0};
	static final double[] truslit_closed = {488, 322.08, 292.8, 390.4, 373.32, 439.2, 923.265, 497.76, 546.56, 570.96, 634.4, 644.16, 614.88, 1113.78, 0.0};
    
	/*
	 * 		
	 * 	double mean = 0;
		for (int i = 0; i < numNotes; i++) {
			double x = noteDur.get(i) - noteAverageDuration;
			x = x / noteStdDeviation;
			
			double y = truslit[i] - truslit_average;
			y = y / truslit_stdev;

			mean += x * y;
		}
		
		double r = (1 / (float)(numNotes - 1)) * mean;

	 */
    private void compareTruslit(double truslit[])
    {
		int numNotes = 14;
		double sum = 0;
		double stddev = 0;
    	for (int i = 0; i < numNotes; i++) {
    		double dur = truslit[i];
			sum += dur;
			stddev += dur * dur;
    	}

		double average = sum / (float)numNotes;
		double truslit_average = average;
		
		average = average * average;
		float n = ((float)numNotes) / (numNotes - 1);
		average = n * average;
		stddev = stddev / (float)(numNotes - 1);
		stddev = (float)Math.sqrt(stddev - average);

		double truslit_stdev = stddev;
    	
		double mean = 0;
		double x_stddev = 0;
		double y_stddev = 0;
		//boolean same = true;
		for (int i = 0; i < numNotes; i++) {
			float x = noteDur.get(i) - noteAverageDuration;
			double y = truslit[i] - truslit_average;
			mean += x * y;
			x_stddev += x * x;
			y_stddev += y * y;
		}
		stddev = Math.sqrt(x_stddev) * Math.sqrt(y_stddev);
		double r = mean / stddev;	
		
		if (r > 1.0) {
			System.out.println("out of bounds " + r + " " + mean);
			for (int i = 0; i < numNotes; i++) {
				System.out.print(noteDur.get(i) + "\t");
			}
			System.out.print("\n");
			for (int i = 0; i < numNotes; i++) {
				System.out.print(truslit[i] + "\t");
			}
			System.out.print("\n");
		}
   	
    	SetSortValue((long) (r * 1000000000f));
    }
    
    public void CompareTruslitOpen()
    {
    	compareTruslit(truslit_open);
    }
	
    public void CompareTruslitClosed()
    {
    	compareTruslit(truslit_closed);
    }

	public void reportSlope(MPOutputStream out)
	{
		out.MPWrite(getUserID() + "\t" + perfDate.getString() + 
				"\t" + getSlope() + "\t" + noteStdDeviation);
		int size = Math.min(getNumNotes(), noteSlope.size());
		for (int j = 0; j < size; j++) {
			out.MPWrite("\t" + noteSlope.get(j));
		}
		out.MPWrite("\t\t" + getLoc().n_lat + " " + getLoc().n_long);
		out.MPWrite("\t" + performanceFile.getName());
		out.MPWriteLn("");	
	}
	
	protected int getSlopeQuadrant(MPSongStats stats)
	{
		if (noteSlope.size() != stats.maxNoteCount) {
			return 0;
		}
		
		int quadrant = 0;
		int numNotes = getNumNotes();
		int quad = numNotes / 4;
		float a[] = { stats.aveNoteSlopeX.get(quad),
				stats.aveNoteSlopeX.get(quad * 2),
				stats.aveNoteSlopeX.get(quad * 3),
				stats.aveNoteSlopeX.get(stats.maxNoteCount - 1) };
		
		float f[] = { noteSlope.get(quad), 
				noteSlope.get(quad * 2), 
				noteSlope.get(quad * 3), 
				noteSlope.get(stats.maxNoteCount - 1)};

		float d[] = { f[0] - a[0], f[1] - a[1], f[2] - a[2], f[3] - a[3] };

		// slow slow
		if (d[0] < 0 && d[3] < 0) {
			quadrant = 1;

			// slow fast
		} else if (d[0] < 0 && d[3] >= 0) {
			quadrant = 2;

			// fast slow
		} else if (d[0] >= 0 && d[3] < 0) {
			quadrant = 3;

			// fast fast
		} else if (d[0] >= 0 && d[3] >= 0) {
			quadrant = 4;
		}
		
		return quadrant;
	}
	
	/**
	 * 
	 * @param phraseindex:  pulled from json, a list of negative and positive indexes into the note array 
	 * indicating the key notes in a phrase
	 * @param stats: the stats of the song (collection of all performances)
	 * @param stdDMult:  a multiplier
	 * @return true if the performance matches the phrase w/i stdDMult
	 */
	protected boolean PhraseMatchSlope(MPPhrase phraseindex, MPSongStats stats)
	{
		boolean match = false;
		boolean negative = false;
		int j = 0;
		double dd = 0;
		do {
			int ci = Math.abs(phraseindex.phrase[j]);
			if (ci >=getNumNotes()) {
				match = false;
			} else {
				// float d1 = p.getNoteDur(ci) - stats.aveNoteDur.get(ci);
				if (phraseindex.phrase[j] < 0) {
					// no double negatives!
					if (negative) {
						match = false;
					} else {
						double d1 = noteSlope.get(ci) - stats.aveNoteSlopeX.get(ci);
						dd += -d1;
						match = d1 < 0;
						negative = match;
					}
				} else {
					double d1 = noteSlope.get(ci) - stats.aveNoteSlopeX.get(ci);
					dd += d1;
					match = d1 > 0;
				}
			}
			++j;
		} while (j < phraseindex.length && match == true);

		if (match) {
			SetSortValue((long)dd);
		}
		
		return match;
	}

	/**
	 * 
	 * @param phraseindex:  pulled from json, a list of negative and positive indexes into the note array 
	 * indicating the key notes in a phrase
	 * @param stats: the stats of the song (collection of all performances)
	 * @param stdDMult:  a multiplier
	 * @return true if the performance matches the phrase w/i stdDMult
	 */
	protected boolean PhraseMatch(MPPhrase phraseindex, MPSongStats stats, double stdDMult)
	{
		boolean match = false;
		boolean negative = false;
		int j = 0;
		double dd = 0;
		do {
			int ci = Math.abs(phraseindex.phrase[j]);
			if (ci >=getNumNotes()) {
				match = false;
			} else {
				// float d1 = p.getNoteDur(ci) - stats.aveNoteDur.get(ci);
				if (phraseindex.phrase[j] < 0) {
					// no double negatives!
					if (false && negative) {
						match = false;
					} else {
						double d1 = getNoteDur(ci) - (stats.aveNoteDur.get(ci) - (stats.stdDevNoteDur.get(ci) * stdDMult));
						dd += -d1;
						match = d1 < 0;
						negative = match;
					}
				} else {
					double d1 = getNoteDur(ci) - (stats.aveNoteDur.get(ci) + (stats.stdDevNoteDur.get(ci) * stdDMult));
					dd += d1;
					match = d1 > 0;
				}
			}
			++j;
		} while (j < phraseindex.length && match == true);

		if (match) {
			SetSortValue((long)dd);
		}
		
		return match;
	}

	protected boolean PhraseMatch(MPPhrase phraseindex, MPSongStats stats)
	{
		// DD!
		return PhraseMatch(phraseindex, stats, 0.0);
	}

	
	/**
	 * Call this method to load meta data from file (if it's not loaded already)
	 * and have it return said meta data. Always call this when accessing meta data
	 * to insure it's loaded from file. Check for null in case meta data does not exist
	 */
	public MPMetaData loadMetaData()
	{
		if (metaData == null) {
			// construct path to file
			File metaFile = new File(performanceFile.getPath().substring(0, performanceFile.getPath().length() - 5) + ".meta");
			
			// load meta data from file
			try {
				FileInputStream fis = new FileInputStream(metaFile);
				ObjectInputStream ois = new ObjectInputStream(fis);
				metaData = (MPMetaData)ois.readObject();
				ois.close();
			} catch (FileNotFoundException e) {
				// do nothing for now
			} catch (IOException e) {
				// do nothing for now
			} catch (ClassNotFoundException e) {
				// do nothing for now
			}
		}

		return metaData;
	}


	/**
	 * This public class encapsulates meta data about the performance such as
	 * information about the user that made the performance. This includes geographic
	 * location (latitude and longitude)
	 */
	public static class MPMetaData implements Serializable
	{
		/** The resource UUID of the actual performance MIDI file on Amazon AWS machine */
		public String resourceUUID;
		
		/** Meta data about performance */
		public String latitude = "";
		public String longitude = "";
		
		/** the unique identifier of the user who created this performances **/
		public int uniqueUserID;
		
		/** the date the performance was made **/
		public Date perfDate;
		
		/** Version UID */
		public static final long serialVersionUID = 1;
		
		/**
		 * Constructor for meta data
		 * @param resourceUUID the resource uuid of the actual performance MIDI file
		 */
		public MPMetaData(String rUUID, int uuID, Date date)
		{
			resourceUUID = rUUID;
			uniqueUserID = uuID;
			perfDate = date;
		}
	}
}

class MPPerformanceDateComparator implements Comparator<MPPerformance> 
{
    public int compare(MPPerformance p1, MPPerformance p2) 
    {
        return p1.compareTo(p2);
    }
}

class MPPerformanceSlopeComparator implements Comparator<MPPerformance> 
{
    public int compare(MPPerformance p1, MPPerformance p2) 
    {
    	int result = Math.round(p1.perfSlope - p2.perfSlope);
    	if (result == 0) {
    		result = p1.compareTo(p2);
    		if (result == 0) {
    			result = p1.getUserID() - p2.getUserID();
    		}
    	}
    	return result;
    }
}

class MPPerformanceUniqueUserComparator implements Comparator<MPPerformance> 
{
    public int compare(MPPerformance p1, MPPerformance p2) 
    {
        return p1.getUserID() - p2.getUserID();
    }
}
	
class MPPerformanceNoteValueComparator implements Comparator<MPPerformance> 
{
    public int compare(MPPerformance p1, MPPerformance p2) 
    {
    	// descending sort order
        return (int)(p2.sortNoteValue - p1.sortNoteValue);
    }
}




