import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collections;
import java.io.File;

/***
 *  
 * @author jeffreysmith
 *
 * This class stores a list of songs (in MPPerformance format) and performs analysis on those songs.
 * 
 * It requires a directory (File class) in the constructor which references a list of files in MIDI format.
 */

public class MPSong {

	protected MPSongList songList;
	
	/** An array of notes hit during performance, ordered chronologically */
	protected MPOriginalPerformance orgPerf;
	protected MPPerfCluster perfCluster;
	protected MPOutputStream out;
	protected MPSongStats stats;
	protected MPSongIO songIO;
	protected MPPhrase expert_phrase;
	protected File song_dir;
	protected int song_perf_max;
			
	public MPSong(MPSongList list, MPOriginalPerformance org, MPOutputStream output)
	{
		stats = new MPSongStats();
		songList = list;
		orgPerf = org;
		out = output;
	}
	
	private void init(File dir, int perf_max, boolean output)
	{
		// instantiate the note event array
		songList = new MPSongList();
		songIO = new MPSongIO(songList, dir, perf_max);

		if (MPAnalyzer.UseOriginalPerformance()) {
			orgPerf = songIO.LoadOrgPerf();
		}
		
		// create the output stream
		if (output) {
			out = new MPOutputStream(songIO.GetSongFileName() + "1_mpdata.xls");
			out.MPWriteLn("\t" + songIO.GetSongFileName());
		}
	}
	
	public MPSong(File dir, int perf_max, boolean output)
	{
		init(dir, perf_max, output);
	}
	
	public MPSong(File dir, int perf_max)
	{
		songList = new MPSongList();
		song_dir = dir;
		song_perf_max = perf_max;
	}
	
	public void LoadMeta()
	{
		init(song_dir, song_perf_max, false);
	}
	
	public void LoadSong()
	{
		stats = new MPSongStats();
		for (MPPerformance perf : songList.songs) {
			if (!perf.LoadPerformance()) {
				System.err.println("song did not load " + GetSongID());
			}
		}
		if (MPAnalyzer.UseOriginalPerformance() && orgPerf != null) {
			if (!orgPerf.LoadPerformance()) {
				System.err.println("song org perf did not load " + GetSongID());
			}
		}
		PrepSongData();
	}
	
	public void OutputSongMetaData()
	{
		init(song_dir, song_perf_max, false);
		for (MPPerformance perf : songList.songs) {
			if (!perf.LoadPerformance()) {
				System.err.println("song did not load " + GetSongID());
			}
			perf.WriteMetaData();
		}
		
	}
	
	public String GetSongID()
	{
		return songIO.songFileName;
	}
	
	public void Close()
	{
		MPGeocoder.WriteGeoCache();	

		if (perfCluster != null) {
			perfCluster.Clear();
		}
		
		if (songList != null) {
			songList.Clear();
		}
		
		if (stats != null) {
			stats.Clear();
		}
		
		if (out != null) {
			out.MPClose();
			out = null;
		}
	}
	
	static int total_songs = 0;
	static int total_retained_songs = 0;

	public void PrepSongData()
	{
		int size = songList.getSize();
		total_songs += size;
		if (MPAnalyzer.ExcludeRepeats()) {
			songList.pruneRepeatPerformances();
		}
		if (MPAnalyzer.ExcludeUrban()) {
			songList.pruneUrban();
		}
		if (MPAnalyzer.ExcludeRural()) {
			songList.pruneRural();
		}
		if (MPAnalyzer.PruneChords()) {
			songList.pruneChords(orgPerf);
		}

		// expert check
		expert_phrase = songIO.getPhraseIndex("cadence");

		// nuke more recent songs as mechanic and game system changed 2/1
		songList.pruneMoreRecent(2013, 2, 1);

		computeStats();
		if (songList.pruneHighStdDevPerfs(stats)) {
			computeStats();
		}
		if (MPAnalyzer.UnpackChords()) {
			stats.calcChords(songList);
			computeStats();
		}
	
		total_retained_songs += songList.getSize();
		System.out.println(total_retained_songs + "/" + total_songs);
	}
		
	public void AnalyzeSongData()
	{
		MPGeocoder.WriteGeoCache();	

		System.out.println(songIO.GetSongFileName()); 

		songList.pruneNonGeoData();
		songList.pruneMoreRecent(2013, 2, 1);
		// songList.pruneRepeatPerformances();
		// songList.pruneFirstPerformances();		
		// songList.pruneUrban();
		computeStats();
		if (songList.pruneHighStdDevPerfs(stats)) {
			computeStats();
		}
	
		songList.pruneChords(orgPerf);
		
		// stats.computeMistakes(songList, orgPerf);
		
		// ReportRepeatPerformances();
		// stats.computeRepeatSongPercent(songList);
		// out.MPWriteLn("\tRepeat Percent" + "\t" + stats.repeatSongPercent);
		
		calcCorrelations();
		AnalyzeByGeo();
		
		stats.ReportNoteStats(out, orgPerf);
		// ReportSlopeByCountry();
	
		// CompareAllPitches();
		
		// AnalyzePhrases();
		// AnalyzePhrases2();
		// AnalyzePhrasesByCountry();
		
		// AnalyzeBySecondPerformance();
		
		AnalyzeByPhraseAndRegion();
		
		// ReportPerformances();

		// AnalyzeOptimalPhrases();
		
		// ReportCorrelations();
		
		// ReportSongSlope();

		// ReportSongInflectionPoints();
		
		// System.out.println("analyze states");
		// AnalyzeByState();
				
		// System.out.println("correlate ");
		// CorrelateByState(MPGeoCoord.getStates());
		// CorrelateByState(MPGeoCoord.getSwingStates());
		// AnalyzeByCountry();
		// CorrelateByCountry();
		
		// ReportDifficulty();
		
		// compareSlopeByQuadrant();
		
		System.out.println("done"); 
	}
	
	public void ReportSlopeByCountry()
	{
		ArrayList<RIRegionItem> countries = songList.CountryListFromPerfList();
		for (RIRegionItem country : countries) {
			MPSongList theperflist = new MPSongList();
			for (MPPerformance perf : songList.songs) {
				if (country.Contains(perf.perfLoc)) {
					theperflist.Add(perf);
				}
			}
			
			if (theperflist.getSize() > 0) {
				// only use the last 200 performances
				MPSongList theperflist2 = new MPSongList();
				for (int i = theperflist.getSize()-1; i >= 0; i--) {
					theperflist2.Add(theperflist.Get(i));
					if (theperflist2.getSize() >= 200) {
						break;
					}
				}
				MPSong song = new MPSong(theperflist2, orgPerf, out);
				song.computeStats();
				out.MPWriteLn(country.GetRegionName() + "\t" + song.songList.getSize() + "\t" + song.stats.averageNoteLength + "\t" + song.stats.stdDevNoteLength);
				song.stats.ReportAveSlope(out);
				out.MPWriteLn("");
			}
		}

		
	}
	
	protected MPPerfQuadrants fetchPerfQuads()
	{
		return songList.fetchPerfQuads(stats);
	}
	
	protected MPRegionDataList fetchRegionData(int region_type)
	{
		MPPhrase phrase = songIO.getPhraseIndex("cadence");
		return songList.fetchRegionData(stats, phrase, region_type);
	}
	
	protected void ReportDifficulty()
	{
		int bichord = orgPerf.CalcDifficulty(1, (long)0);
		int trichord = orgPerf.CalcDifficulty(2, (long)0);
		int polychord = orgPerf.CalcDifficulty(3, 9, (long)0);
				
		int easy = 0;
		int med = 0;
		int hard = 0;
		
		for (MPPerformance p : songList.songs) {
			int c1 = p.CalcDifficulty(1, (long)0);
			int c2 = p.CalcDifficulty(2, (long)0);
			int c3 = p.CalcDifficulty(3, 9, (long)0);
			if (c1 == bichord && c2 == trichord && c3 == polychord) {
				++easy;
				System.out.println("easy " + c1 + " " + c2 + " " + c3 + " " + p.getUserID() + " " + p.perfLoc.address);
			} else if (c1 == (bichord + trichord) && c2 == polychord && c3 == 0) { 
				++med;
				System.out.println("med " + c1 + " " + c2 + " " + c3 + " " + p.getUserID() + " " + p.perfLoc.address);
			} else {
				++hard;
				System.out.println("hard " + c1 + " " + c2 + " " + c3 + " " + p.getUserID() + " " + p.perfLoc.address);
			}
		}
		
		System.out.println("org ");
		System.out.println(bichord + " " + trichord + " " + polychord + " ");
		System.out.println("easy med hard ");
		System.out.println(easy + " " + med + " " + hard);
	}
	
	protected void CompareAllPitches()
	{
		for (MPPerformance p : songList.songs) {
			for (int i = 0; i < Math.min(p.midi.noteEvents.size(), orgPerf.midi.noteEvents.size()); i++) {
				String pitch1 = p.midi.noteEvents.get(i).getPitch();
				String pitch2 = p.midi.noteEvents.get(i).getPitch();
				if (pitch1.compareTo(pitch2) != 0) {
					System.out.println(i + " " + pitch1 + "\t" + pitch2);
				}
			}
		}		
	}

	protected MPSongList NewSongListFromPhraseSlope(MPPhrase phraseindex, double stdDMult) 
	{
		MPSongList hits = new MPSongList();
		
		// find performances above average (or below if negative phraseindex) on all phraseindex notes
		for (MPPerformance p : songList.songs) {
			
			boolean match = p.PhraseMatchSlope(phraseindex, stats);
			if (match) {
				// System.out.println("hit " + p.uniqueUserID + " " + dd + " stddev " + stats.stdDevNoteLength);
				hits.Add(p);
			}
		}
		
		// sort by a weighting of how well this a performance met the phrase criteria
		hits.SortByValue();
			
		return hits;
	}
	
	class SongLists {
		SongLists() 
		{
			experts = new MPSongList();
			amateurs = new MPSongList();
		}
		
		MPSongList experts;
		MPSongList amateurs;
	}

	private SongLists NewSongListsFromPhrase(MPPhrase phraseindex, double stdDMult) 
	{
		SongLists results = new SongLists();
		
		// find performances above average (or below if negative phraseindex) on all phraseindex notes
		for (MPPerformance p : songList.songs) {
			
			boolean match = p.PhraseMatch(phraseindex, stats, stdDMult);
			if (match) {
				// System.out.println("hit " + p.uniqueUserID + " " + dd + " stddev " + stats.stdDevNoteLength);
				results.experts.Add(p);
			} else {
				results.amateurs.Add(p);
			}
		}
		
		// sort by a weighting of how well this a performance met the phrase criteria
		results.experts.SortByValue();
			
		return results;
	}

	private MPSongList NewSongListFromPhrase(MPPhrase phraseindex, double stdDMult) 
	{
		MPSongList hits = new MPSongList();
		
		// find performances above average (or below if negative phraseindex) on all phraseindex notes
		for (MPPerformance p : songList.songs) {
			
			boolean match = p.PhraseMatch(phraseindex, stats, stdDMult);
			if (match) {
				// System.out.println("hit " + p.uniqueUserID + " " + dd + " stddev " + stats.stdDevNoteLength);
				hits.Add(p);
			}
		}
		
		// sort by a weighting of how well this a performance met the phrase criteria
		hits.SortByValue();
			
		return hits;
		
	}
	/***
	 * Construct a new song list from a phrase
	 * 
	 */
	protected MPSongList NewSongListFromPhrase(MPPhrase phraseindex)
	{
		double d =        0.01;
		double stdDMult = 0.5; // stats.stdDevNoteLength / stats.averageNoteLength;
		
		MPSongList l = null;
		do { 
			l = NewSongListFromPhrase(phraseindex, stdDMult);
			if (l != null && l.getSize() > 0) {
				break;
			}
			// System.out.println("miss " + stdDMult);
			stdDMult -= d;	
		} while (true);
		
		System.out.println("match " + stdDMult);
		
		return l;
	}
	
	/***
	 * 
	 * look for songs where the average duration of phrase boundaries is above average
	 * 
	 * return a new MPSong class with such performances
	 */

	private MPSong NewSongFromPerfList(MPSongList perflist)
	{
		if (perflist == null || perflist.getSize() == 0) {
			return null;
		}
		
		MPSong newSong = new MPSong(perflist, orgPerf, out);
		
		newSong.computeStats();
		
		newSong.perfCluster = perfCluster;
		newSong.songIO = songIO;
		newSong.out = out;

		return newSong;
	}
	
	private void addPhrase(MPOptimalPhraseList list, MPPhrase thephrase, int k, int depth, int mincluster)
	{
		thephrase.set(depth, k);

		float themult = (float)(stats.stdDevNoteLength / stats.averageNoteLength) ; // * 5/4;
		MPSongList l = NewSongListFromPhrase(thephrase, themult);
		if (l != null && l.getSize() > mincluster) {
			System.out.println(thephrase.GetString());
			if (perfCluster.CalcAvePearson(l)) {
				list.AddItem(thephrase, perfCluster.GetData().pearson, l.getSize());
			}
		}
				
	}

	private void phraseSearch(MPOptimalPhraseList list, MPPhrase thephrase, int k, int depth, int maxdepth, int mincluster)
	{
		if (depth == maxdepth) {
			return;
		}
		
		if (k == stats.maxNoteCount) {
		// if (k == 250) {
			return;
		}
		
		addPhrase(list, thephrase, k, depth, mincluster);
		phraseSearch(list, thephrase, k + 1, depth + 1, maxdepth, mincluster);
		phraseSearch(list, thephrase, k + 1, depth, maxdepth, mincluster);
	}

	protected void AnalyzeOptimalPhrases()
	{		
		int minclustersize = songList.getSize() / 5;
		int maxdepth = 4;
		MPPhrase phraseIndex = new MPPhrase(maxdepth);
		
		MPOptimalPhraseList list = new MPOptimalPhraseList();

		// maxdepth indexes
		// phraseSearch(list, phraseIndex, - stats.maxNoteCount / 30, 0, maxdepth, minclustersize);
		phraseSearch(list, phraseIndex, -100, 0, maxdepth, minclustersize);
		
		System.out.println("sort");
		list.Sort();

		System.out.println("output");
		list.Output(out);
	}
	
	protected void AnalyzePhrases2()
	{					
		float stdmult = (stats.stdDevNoteLength / stats.averageNoteLength) / 4;
		MPPhrase pi1 = songIO.getPhraseIndex("cadence");
		MPSong ns = NewSongFromPerfList(NewSongListFromPhrase(pi1, stdmult));
		if (ns == null) {
			return;
		}
		ns.stats.ReportAveSlope(out);
		
		MPPhrase pi2 = songIO.getPhraseIndex("cadence2");
		MPSong ns2 = NewSongFromPerfList(NewSongListFromPhrase(pi2, stdmult));
		if (ns2 == null) {
			return;
		}
		
		/* find intersection
		System.out.println("ns2 size pre " + ns2.songList.size());
		for (MPPerformance p : ns.songList) {
			if (ns2.perfExists(p)) {
				ns2.songList.remove(p);
			}
		}
		System.out.println("ns2 size post " + ns2.songList.size());
		*/

		ns2.stats.ReportAveSlope(out);
				
		out.MPWriteLn("");
		out.MPWriteLn("phrase cluster 1");
		out.MPWrite("\tcluster rank\t" + ns.perfCluster.CalcAvePearson(ns.songList) + "\t");
		out.MPWrite(pi1.GetString());
		out.MPWriteLn(perfCluster.CalcAvePearson(songList) + " ");
		ns.AnalyzeByGeo();
		ns.ReportPerformances();

		// ns.perfCluster.ReportClusterByIndex(out, ns.songList, 400);
		
		// ns.perfCluster.Report(out, ns.songList, 5);

		// ns.perfCluster.Report(out, ns.songList);
		// ns.ReportSongSlope(out, ns.songList);
		
		out.MPWriteLn("");
		out.MPWriteLn("phrase cluster 2");
		out.MPWrite("\tcluster rank\t" + ns2.perfCluster.CalcAvePearson(ns2.songList) + "\t");
		out.MPWrite(pi2.GetString());
		out.MPWriteLn(perfCluster.CalcAvePearson(songList) + " ");

		ns2.AnalyzeByGeo();
		ns2.ReportPerformances();
		
		// ns2.perfCluster.ReportClusterByIndex(out, ns2.songList, 400);
		// ns2.perfCluster.Report(out,  ns2.songList, 5);
	}

	protected void AnalyzeBySecondPerformance()
	{
		Collections.sort(songList.songs);
		int numrepeats = 3;
		MPSongList sl[] = new MPSongList[numrepeats];
		for (int i = 0; i < numrepeats; i++) {
			sl[i] = new MPSongList();
		}

		int i = 0; 
		while (i < songList.getSize() - 1) {
			MPPerformance a = songList.Get(i);
			int j = i + 1;
			boolean repeat = true;
			while (repeat && j < songList.getSize()) {
				MPPerformance b = songList.Get(j);
				if (a.getUserID() == b.getUserID()) {
					// first repeat?
					if (j == i + numrepeats - 1) {
						for (int k = 0; k < numrepeats; k++) {
							sl[k].Add(songList.Get(i + k));
						}
					}
					++j;
				} else {
					repeat = false;
				}
			}
			i = j;
		}		
		
		String s = "original size, 1st repeat size " + songList.getSize();
		for (i = 0; i < numrepeats; i++) {
			s += ", " + sl[i].getSize();
		}
		
		System.out.println(s);
		
		for (i = 0; i < numrepeats; i++) {
			MPSong s1 = NewSongFromPerfList(sl[i]);
			out.MPWriteLn("\nRepeats " + i );
			s1.AnalyzeByGeo();
			s1.stats.ReportNoteStats(out, orgPerf);
			s1.ReportPerformances();
		}
	}

	protected void AnalyzeByPhraseAndRegion()
	{		
		SongLists lists = NewSongListsFromPhrase(songIO.getPhraseIndex("cadence"), 0.3);
		// MPSongList cadence1 = NewSongListFromPhraseSlope(songIO.getPhraseIndex("cadence"), 0.0);
		
		if (lists.experts.getSize() == 0) {
			return;
		}
		
		// only keep performances above average tempo?
		boolean testtempo = false;
		if (testtempo) {
			MPSongList experts2 = new MPSongList();
			for (MPPerformance i : lists.experts.songs) {
				float x = i.getNoteSlope(i.getNumNotes() - 1);
				float y = 0 ; // stats.aveNoteSlopeX.get(42);
				if (x >= y) {
					experts2.Add(i);
				} else {
					lists.amateurs.Add(i);					
				}
			}
			lists.experts = experts2;
		}
		
		if (lists.experts.getSize() == 0) {
			return;
		}
		
		System.out.println("number of experts " + lists.experts.getSize() + " " + lists.amateurs.getSize());
		MPSong c1 = NewSongFromPerfList(lists.experts);
		if (c1 != null) {
			out.MPWriteLn("\nexperts\t");
			c1.AnalyzeByGeo();
			c1.ReportNoteStats(out);
			c1.ReportPerformances();
		}

		MPSong c2 = NewSongFromPerfList(lists.amateurs);
		if (c2 != null) {
			out.MPWriteLn("\namateurs\t");
			c2.AnalyzeByGeo();
			c2.ReportNoteStats(out);
			c2.ReportPerformances();
		}
	}

	protected void ReportNoteStats(MPOutputStream output)
	{
		stats.ReportNoteStats(output, orgPerf);
	}
	
	protected void AnalyzePhrasesByTruslit()
	{			
		for (MPPerformance p : songList.songs) {
			p.CompareTruslitOpen();
		}
		songList.SortByValue();
		MPSongList openlist = new MPSongList();
		
		// add all performances with correlation great than .9 to list
		for (int i = 0; i < songList.getSize(); i++) {
			MPPerformance p = songList.Get(i);
			if (p.GetSortValue() >= .7 * 1000000000f) {
				openlist.Add(p);
			} else {
				break;
			}
		}
		
		if (openlist.getSize() > 0) {
			MPSong c1 = NewSongFromPerfList(openlist);
			c1.perfCluster.CalcAvePearson(openlist);
			out.MPWriteLn("\ntruslit open");
			c1.AnalyzeByGeo();
			c1.stats.ReportNoteStats(out, orgPerf);
			c1.ReportPerformances();
		}
		
		for (MPPerformance p : songList.songs) {
			p.CompareTruslitClosed();
		}
		songList.SortByValue();
		MPSongList closedlist = new MPSongList();
		for (int i = 0; i < songList.getSize(); i++) {
			MPPerformance p = songList.Get(i);
			if (p.GetSortValue() >= .7 * 1000000000f) {
				closedlist.Add(p);
			} else {
				break;
			}
		}

		if (closedlist.getSize() > 0) {
			MPSong c2 = NewSongFromPerfList(closedlist);
			c2.perfCluster.CalcAvePearson(closedlist);
			out.MPWriteLn("\ntruslit closed");
			c2.AnalyzeByGeo();
			c2.stats.ReportNoteStats(out, orgPerf);
			c2.ReportPerformances();
		}
	}
	
	/**
	 * 
	 * look at how song experts compare by country
	 * 
	 */
	protected void AnalyzePhrasesByCountry()
	{					
		MPPhrase pi1 = songIO.getPhraseIndex("cadence");
		if (pi1 == null) {
			return;
		}
		
		MPSongList newperflist = NewSongListFromPhrase(pi1, 0.0);
		MPSong ns = NewSongFromPerfList(newperflist);
		if (ns == null) {
			return;
		}
		
		/*
		 * TBD  move to new tree structure
		 * 
		MPRegionPairList rl = CorrelateByCountry(null);
		rl.AggregateAndReport(out, fetchUniqueIDs(), fetchRegionData(MPAnalyzer.ANALYZE_BY_COUNTRY), 10);
		
		MPRegionPairList rlcntry = ns.CorrelateByCountry(null);
		// use original song stats; otherwise expert/phrase calc is skewed
		ns.stats = stats;
		rlcntry.AggregateAndReport(out, ns.fetchUniqueIDs(), ns.fetchRegionData(MPAnalyzer.ANALYZE_BY_COUNTRY), 10);
		*/
	}

	
	protected void AnalyzePhrases()
	{					
		MPPhrase pi1 = songIO.getPhraseIndex("cadence");
		if (pi1 == null) {
			return;
		}
		
		MPSongList newperflist = NewSongListFromPhrase(pi1, 0.0);
		MPSong ns = NewSongFromPerfList(newperflist);
		if (ns == null) {
			return;
		}
				
		out.MPWriteLn("");
		out.MPWriteLn("phrase cluster 1\t" + newperflist.getSize());
		out.MPWrite("\tcluster rank\t" + ns.perfCluster.CalcAvePearson(ns.songList) + "\t");
		out.MPWrite(pi1.GetString());
		out.MPWriteLn(perfCluster.CalcAvePearson(songList) + " ");
		String s = "\turban %\t" + songList.urbanPercent() + "\t";
		s = s + "expert %\t" + (float)ns.songList.getSize() / (float)songList.getSize() + "\t";
		s = s + "expert urban %\t" + ns.songList.urbanPercent();
		System.out.println(s);
		out.MPWriteLn(s);
		
		out.MPWriteLn("");
		// ns.perfCluster.RankandReport(out, ns.songList);
	}
	
	/*
	 * this version of AnalyzePhrases worked OK for nearer my god but failed miserably on Amazing Grace
	 * 
	 * It picked a song that had a huge delay on every cadence and was anomalous
	 * 
	 *
	protected void AnalyzePhrases(MPOutputStream out)
	{					
		MPPhrase pi1 = songIO.getPhraseIndex("cadence");
		MPSong ns = NewSongFromPhrase(NewSongListFromPhrase(pi1));
		if (ns == null) {
			return;
		}
		
		ns.stats.ReportAveSlope(out);
		
		out.MPWriteLn("");
		out.MPWriteLn("phrase cluster 1");
		out.MPWrite("\tcluster rank\t" + ns.perfCluster.GetAveSpearmanIndex() + "\t");
		out.MPWrite(pi1.GetString());

		out.MPWriteLn(perfCluster.GetAveSpearmanIndex() + " ");
		ns.AnalyzeByGeo(out, ns.songList);
		
		// ns.perfCluster.SortBySpearman();
		// ns.perfCluster.Report(out, ns.songList, 250);

		out.MPWriteLn("");
		out.MPWriteLn("");
		out.MPWriteLn("");
		
		// find 'most representative song' from this cluster
		Collections.sort(ns.songList, new MPPerformanceNoteValueComparator());
		ns.ReportPerformances(out);
		out.MPWriteLn("");
		out.MPWriteLn("");

		MPPerformance p = ns.songList.get(0);
		
		// find index of this performance in parent cluster
		int index = -1;
		for (int i = 0; i < songList.size(); i++) {
			MPPerformance a = songList.get(i);
			if (a.equals(p)) {
				index = i;
				break;
			}
		}
		
		// Q:  what is the difference between top 100 relatives to best song vs top 100 matches of phrase?  LISTEN!
		// Q:  how do we find velocity and arch of velocity?
		
		// now find relatives searching original cluster
		MPPerfCluster relatives = perfCluster.FindAllRelatives(index);
		
		// output top 250 sorted by rank
		relatives.SortBySpearman();
		relatives.Report(out, songList, 250);
	}
	*/
	
	private void computeStats()
	{
		if (songList.getSize() < 1) {
			return;
		}
		
		stats.computeStats(songList);

		if (MPAnalyzer.UseOriginalPerformance()) {
			stats.computeAveSlope(songList, calcBaselineofPerformance(orgPerf));
		} else {
			stats.computeAveSlope(songList, stats.aveNoteDur);
		}
		
		// stats.computeAveInflection(songList);
	}
	
	/***
	 * calc pearson and spearman correlation coefficients
	 */
	
	protected void calcCorrelations()
	{
		perfCluster = new MPPerfCluster(songList);
	}
	
	public void ReportRepeatPerformances()
	{
		songList.Sort();

		MPOutputStream repeatOut = new MPOutputStream(songIO.GetSongFileName() + "_repeat_mpdata.xls");
		
		repeatOut.MPWriteLn("\t" + songIO.GetSongFileName());

		int i = 0; 
		while (i < songList.getSize() - 1) {
			MPPerformance a = songList.Get(i);
			repeatOut.MPWrite(a.getUserID() + "\t" + MPGeoCoord2.geoIndexToString(a.perfLoc.geoIndex()) + "\t" + a.noteAverageDuration);

			int j = i + 1;
			boolean repeat = true;
			while (repeat && j < songList.getSize()) {
				MPPerformance b = songList.Get(j);
				if (a.getUserID() == b.getUserID()) {
					repeatOut.MPWrite("\t" + b.noteAverageDuration);
					++j;
				} else {
					repeat = false;
				}
			}
			i = j;
			repeatOut.MPWriteLn("");
		}		

		repeatOut.MPClose();
//		System.out.println("total repeat count " + repeatcount + " " + songList.size()); 

	}
	
	public void AnalyzeByHour()
	{
		int songHours[] = {0, 0, 0, 0};
		for (MPPerformance aPerformance : songList.songs) {
			songHours[aPerformance.perfDate.hourIndex()]++;
		}
		
		out.MPWriteLn("\tanalyze by hour");
		
		for (int i = 0; i < 4; i++) {
			out.MPWrite(MPPerfDate.hourIndexToString(i) + "\t" + songHours[i] + "\t");
		}
		out.MPWriteLn("");
	}
	
	/*
		
	public void DivideByGeo(ArrayList<MPPerformance> list, ArrayList<MPSong> songs)
	{
		for (MPPerformance aPerformance : list) {
			short index = aPerformance.perfLoc.geoIndex();
			ave[index] += aPerformance.averageDuration;

	}
	*/
	
	public void AnalyzeByGeo()
	{
		out.MPWriteLn("");
		
		// float ave_n = new Float(list.size());
		float ave[] = {0, 0, 0, 0, 0, 0};
		float stddev[] = {0, 0, 0, 0, 0, 0};
		short ave_c[] = {0, 0, 0, 0, 0, 0};
		short urban_c[] = {0, 0, 0, 0, 0, 0};
		
		for (MPPerformance aPerformance : songList.songs) {
			if (!aPerformance.perfLoc.isGeoCodeNull()) {
				short index = aPerformance.perfLoc.geoIndex();
				ave[index] += aPerformance.noteAverageDuration;
				stddev[index] += aPerformance.noteAverageDuration * aPerformance.noteAverageDuration;
				ave_c[index]++;
				if (aPerformance.perfLoc.is_city){
					urban_c[index]++;
				}
			}
		}
		
		for (int i = 0; i < 6; i++) {
			stddev[i] = stddev[i] / ave_c[i];
			ave[i] = ave[i] / ave_c[i];
			stddev[i] = stddev[i] - (ave[i] * ave[i]);
			stddev[i] = (float)Math.sqrt((double)stddev[i]);
		}
		
		out.MPWriteLn("geo\tcount \tave \tstddev \torgsong ave \texperts% \turban% \tpearson");
		out.MPWrite("world\t" + songList.getSize() + "\t" + stats.averageNoteLength + "\t" + stats.stdDevNoteLength + "\t");
		if (orgPerf != null) {
			out.MPWrite("" + orgPerf.noteAverageDuration + "\t");
		} else {
			out.MPWrite("0.0\t");
		}
		
		float experts = songList.GetExperts(songIO.getPhraseIndex("cadence"), stats);
		out.MPWrite((experts / (float)songList.getSize()) + "\t" + songList.urbanPercent() + "\t");
		
		// this is expensive
		if (songList.getSize() <= 200000) {
			perfCluster.CalcAvePearson(songList);
			System.out.println("pearson " + perfCluster.clusterData.pearson);
			out.MPWriteLn(perfCluster.clusterData.pearson);
		}

		out.MPWriteLn("\n\n\tcount \tave \tstddev \turban percent");
		
		for (int i = 0; i < 6; i++) {
			out.MPWriteLn(MPGeoCoord2.geoIndexToString(i) + "\t" + ave_c[i] + "\t" + ave[i]+ "\t" + stddev[i] + "\t" + (float)urban_c[i]/(float)ave_c[i]);
		}
		
		/*
		GRMapData data = new GRMapData();
		data.LoadGraph();
		Graph(data, 0.6f, 1.0f);
		data.StoreGraph();
		*/
	}
	
	public void Graph(GRMapData data, float radius, float color)
	{
		for (MPPerformance aPerformance : songList.songs) {
			String s = aPerformance.perfLoc.n_lat + "," + aPerformance.perfLoc.n_long;
			data.DrawPoint(s, radius, color);
		}
	}
	
	// factor this!!!  parameters:  string list and callback boolean 
	
	public void AnalyzeByCountry()
	{
		float ave[] = {0, 0};
		float stddev[] = {0, 0};
		short ave_c[] = {0, 0};

		MPSongList us = new MPSongList();
		MPSongList ca = new MPSongList();
		for (MPPerformance perf : songList.songs) {
			int index = -1;
			if (perf.perfLoc.isUS() || perf.perfLoc.isCanada()) {
				if (perf.perfLoc.isUS()) {
					index = 0;
					us.Add(perf);
				} else {
					index = 1;
					ca.Add(perf);
				}
				ave[index] += perf.noteAverageDuration;
				stddev[index] += perf.noteAverageDuration * perf.noteAverageDuration;
				ave_c[index]++;
			}
		}
		
		if (ca.getSize() < 1 || us.getSize() < 1) {
			return;
		}

		for (int i = 0; i < 2; i++) {
			stddev[i] = stddev[i] / ave_c[i];
			ave[i] = ave[i] / ave_c[i];
			stddev[i] = stddev[i] - (ave[i] * ave[i]);
			stddev[i] = (float)Math.sqrt((double)stddev[i]);
		}

		// find world experts
		MPSongList experts = NewSongListFromPhrase(songIO.getPhraseIndex("cadence"), 0.0);

		out.MPWriteLn("");
		out.MPWriteLn("");
		out.MPWriteLn("geo\tcount \tave \tstddev \tpearson \texperts \tsonglength \texpert ratio");
		out.MPWriteLn("world\t" + songList.getSize() + "\t" + stats.averageNoteLength + "\t" + stats.stdDevNoteLength + "\t" + 
				perfCluster.CalcAvePearson(songList) + "\t" + experts.getSize() + "\t" + stats.averageNoteLength * stats.maxNoteCount + 
				"\t" + (((float)experts.getSize())/songList.getSize()));
		out.MPWriteLn("");

		MPSong ussong = NewSongFromPerfList(us);
		MPSong casong = NewSongFromPerfList(ca);

		int usexperts = experts.perfIntersectionCnt(us);
		int caexperts = experts.perfIntersectionCnt(ca);

		int i = 0;
		out.MPWriteLn("US"+ "\t" + ave_c[i] + "\t" + ave[i]+ "\t" + stddev[i] + "\t" + 
				perfCluster.CalcAvePearson(us) + "\t" + usexperts + "\t" + ussong.stats.averageNoteLength * ussong.stats.maxNoteCount + 
				"\t" + ((float)usexperts / ave_c[i]));

		i = 1;
		out.MPWriteLn("CA"+ "\t" + ave_c[i] + "\t" + ave[i]+ "\t" + stddev[i] + "\t" + 
				perfCluster.CalcAvePearson(ca) + "\t" + caexperts + "\t" + casong.stats.averageNoteLength * casong.stats.maxNoteCount + 
				"\t" + ((float)caexperts / ave_c[i]));
	}

	public void CorrelateInCountry(ArrayList<RIRegionItem> countryList)
	{
		for (int i = 0; i < countryList.size(); i++) {
			RIRegionItem icountry = countryList.get(i);
			MPSongList theperflist1 = songList.getSongsInRegion(icountry);
			int isize = theperflist1.getSize();
			if (isize > 0) {
				icountry.AddRegionUserData(theperflist1, expert_phrase, stats);
				if (isize > 1) {
					if (perfCluster.CalcAvePearson(theperflist1)) {
						icountry.AddIntraRegionItems(perfCluster.GetData());
					}
					theperflist1.songs.clear();
				}
			}
		}
	}
	
	/***
	 * 
	 * test affinity by country to country across a list of songs for a given countryList
	 * 
	 * NOTE:  calcCorrelations must be called prior, as this function has a dependency
	 * on the cluster arrays.
	 * 
	 * @param list
	 * @param countryList
	 */
	
	public void CorrelateToRegion(ArrayList<RIRegionItem> compare_regions, ArrayList<RIRegionItem> region_item_list)
	{
		MPPerfCluster the_cluster = new MPPerfCluster(songList);
		for (RIRegionItem the_region : compare_regions) {
			// add all of the perfs in this region of this song
			MPSongList regionperflist = the_region.getSongsInRegion(GetSongID());
			
			// no perfs of this region in this song?
			if (regionperflist.getSize() > 0) {
			
				// this isn't working
				// the_region.AddRegionUserData(regionperflist, expert_phrase, stats);
				
				// traverse regionperflist and compare each to the_region
				for (RIRegionItem ri : region_item_list) {
					if (!ri.equals(the_region)) {
						MPSongList ri_regionperflist = ri.getSongsInRegion(GetSongID());
			
						// calc region-to-region  
						if (ri_regionperflist.getSize() > 0) {
							if (the_cluster.CalcAvePearson(regionperflist, ri_regionperflist)) {
								the_region.AddInterRegionItems(ri, the_cluster.GetData());
								ri.AddInterRegionItems(the_region, the_cluster.GetData());
							}
						}
						ri_regionperflist.songs.clear();
					}
				}
		
				regionperflist.songs.clear();
			}
		}
		the_cluster.Clear();
	}
	
	public void CorrelateByRegion(ArrayList<RIRegionItem> regionitems) 
	{
		boolean debug_time = false;
		boolean calc_intra = true;
		boolean calc_inter = false;
		long chunks = regionitems.size() / 10;
		long cur_time = Calendar.getInstance().getTimeInMillis();
		
		boolean report_region_songs = false;
		MPOutputStream out2 = null;
		if (report_region_songs) {
			out2 = new MPOutputStream("region_list.xls");
		}

		MPPerfCluster the_cluster = new MPPerfCluster(songList);
		for (int i = 0; i < regionitems.size(); i++) {
			RIRegionItem i_region = regionitems.get(i);
			MPSongList i_perflist = i_region.getSongsInRegion(GetSongID());
			int i_size = i_perflist.getSize();
			
			if (debug_time && i > 0 && i % chunks == 0) {
				System.out.println((i / (float)regionitems.size()) + "% " + 
						(Calendar.getInstance().getTimeInMillis() - cur_time) / (float)i);
			}
			
			if (i_size > 0) {
				
				i_region.AddRegionUserData(i_perflist, expert_phrase, stats);
				
				if (report_region_songs) {
					MPSong temp = NewSongFromPerfList(i_perflist);
					out2.MPWriteLn(i_region.GetFullName());
					temp.ReportNoteStats(out2);
					out2.MPWriteLn("");
					// temp.ReportPerformances(output);
					// temp.ReportExpertPerformances(output, expert_phrase, stats);
					// int pre_size = i_perflist.getSize();
					// i_perflist.pruneChords(orgPerf);
					// out3.MPWriteLn(i_region.GetFullName() + "\t" + i_perflist.getSize() + "\t" + pre_size);
				}
				
				// calc region i (intra)
				if (calc_intra && i_size > 1) {
					if (the_cluster.CalcAvePearson(i_perflist)) {
						i_region.AddIntraRegionItems(the_cluster.GetData());
					}
				}

				if (calc_inter) {
					for (int j = i + 1; j < regionitems.size(); j++) {
						RIRegionItem j_region = regionitems.get(j);
						MPSongList j_perflist = j_region.getSongsInRegion(GetSongID());
						int j_size = j_perflist.getSize();
			
						// calc region-to-region (inter)
						if (j_size > 0) {
							if (the_cluster.CalcAvePearson(i_perflist, j_perflist)) {
								i_region.AddInterRegionItems(j_region, the_cluster.GetData());
								j_region.AddInterRegionItems(i_region, the_cluster.GetData());
							}
						}
				
						j_perflist.songs.clear();
					} // j
				}
			} // if (isize > 0)
			i_perflist.songs.clear();
		} // i
		
		if (out2 != null) {
			out2.MPClose();
		}
		the_cluster.Clear();
	}
	
	public void CorrelateByState(String[] statelist) 
	{
		int redcount = 0;
		int bluecount = 0;

		for (String s : statelist) {
			MPSongList red = new MPSongList();
			MPSongList blue = new MPSongList();
			
			// walk the list of all performances for a given song
			for (MPPerformance perf : songList.songs) {
				if (!perf.perfLoc.isGeoCodeNull() && perf.perfLoc.isUS()) {
					if (s.equalsIgnoreCase(perf.perfLoc.getState())) {
						red.Add(perf);
						blue.Add(perf);
					} else if (perf.perfLoc.isRedState()) {
						red.Add(perf);
					} else {
						blue.Add(perf);
					}
				}
			}
			
			float aveRed = (float)perfCluster.CalcAveSpearman(red);
			float aveBlue = (float)perfCluster.CalcAveSpearman(blue);
			
			if (aveRed == aveBlue) {
				out.MPWriteLn(s + " is split");
			} else if (aveRed > aveBlue) {
				out.MPWriteLn(s + "\t" + aveRed + "\t" + aveBlue + "\tis red");
				++redcount;
			} else {
				out.MPWriteLn(s + "\t" + aveRed + "\t" + aveBlue + "\tis blue");
				++bluecount;
			}
		}
		
		System.out.println("red " + redcount + " blue " + bluecount);
	}	

	public void AnalyzeByState() 
	{
		float ave[] = {0, 0};
		float stddev[] = {0, 0};
		short ave_c[] = {0, 0};
		
		MPSongList red = new MPSongList();
		MPSongList blue = new MPSongList();
		MPSongList swing = new MPSongList();
		
		for (MPPerformance perf : songList.songs) {
			if (!perf.perfLoc.isGeoCodeNull() && perf.perfLoc.isUS()) {
				int index = perf.perfLoc.isRedState() ? 0 : 1;
				if (index == 0) {
					red.Add(perf);
				} else {
					blue.Add(perf);
				}
				if (perf.perfLoc.isSwingState()) {
					swing.Add(perf);
				}
				
				ave[index] += perf.noteAverageDuration;
				stddev[index] += perf.noteAverageDuration * perf.noteAverageDuration;
				ave_c[index]++;
			}
		}

		for (int i = 0; i < 2; i++) {
			stddev[i] = stddev[i] / ave_c[i];
			ave[i] = ave[i] / ave_c[i];
			stddev[i] = stddev[i] - (ave[i] * ave[i]);
			stddev[i] = (float)Math.sqrt((double)stddev[i]);
		}
	
		MPSongList experts = NewSongListFromPhrase(songIO.getPhraseIndex("cadence"), 0.0);
		
		out.MPWriteLn("");
		out.MPWriteLn("");
		out.MPWriteLn("geo\tcount \tave \tstddev \tpearson \texperts \tsonglength \texpert ratio");
		out.MPWriteLn("world\t" + songList.getSize() + "\t" + stats.averageNoteLength + "\t" + stats.stdDevNoteLength + "\t" + 
				perfCluster.CalcAveSpearman(songList) + "\t" + experts.getSize() + "\t" + stats.averageNoteLength * stats.maxNoteCount + 
				"\t" + (((float)experts.getSize())/songList.getSize()));
		out.MPWriteLn("");
		
		MPSong redsong = NewSongFromPerfList(red);
		MPSong bluesong = NewSongFromPerfList(red);
		MPSong swingsong = NewSongFromPerfList(red);
		int redexperts = experts.perfIntersectionCnt(red);
		int blueexperts = experts.perfIntersectionCnt(blue);
		int swingexperts = experts.perfIntersectionCnt(swing);
		
		int i = 0;
		out.MPWriteLn("red"+ "\t" + ave_c[i] + "\t" + ave[i]+ "\t" + stddev[i] + "\t" + 
				perfCluster.CalcAveSpearman(red) + "\t" + redexperts + "\t" + redsong.stats.averageNoteLength * redsong.stats.maxNoteCount + 
				"\t" + ((float)redexperts / red.getSize()));
		
		i = 1;
		out.MPWriteLn("blue"+ "\t" + ave_c[i] + "\t" + ave[i]+ "\t" + stddev[i] + "\t" + 
				perfCluster.CalcAveSpearman(blue) + "\t" + blueexperts + "\t" + bluesong.stats.averageNoteLength * bluesong.stats.maxNoteCount + 
				"\t" + ((float)blueexperts / blue.getSize()));
		
		out.MPWriteLn("swing" + "\t\t\t\t" + swingsong.perfCluster.CalcAveSpearman(songList) + "\t" + swingexperts + "\t" + swingsong.stats.averageNoteLength * swingsong.stats.maxNoteCount +
				"\t" + ((float)swingexperts / swing.getSize()));
		out.MPWriteLn("");
		
		// compareSlopeByQuadrant();
		// redsong.compareSlopeByQuadrant();
		// bluesong.compareSlopeByQuadrant();
		
		// state my state
		out.MPWriteLn("Red");
		outputStates(redsong, MPGeoCoord2.getRedStates(), experts);
		out.MPWriteLn("Blue");
		outputStates(bluesong, MPGeoCoord2.getBlueStates(), experts);
		out.MPWriteLn("Swing");
		outputStates(swingsong, MPGeoCoord2.getSwingStates(), experts);
		
	}
	
	private void outputStates(MPSong song, String[] statelist, MPSongList experts) 
	{
		for (String s : statelist) {
			MPSongList state = new MPSongList();
			for (MPPerformance p : song.songList.songs) {
				if (p.perfLoc.isUSState(s)) {
					state.Add(p);
				}
			}
			if (state.getSize() > 0) {
				MPSong statesong = NewSongFromPerfList(state);
				int expertcnt = experts.perfIntersectionCnt(state);
				out.MPWriteLn(s + "\t" + statesong.songList.getSize() + "\t" + statesong.stats.averageNoteLength + "\t" + 
							statesong.stats.stdDevNoteLength + "\t" + statesong.perfCluster.CalcAvePearson(statesong.songList) + "\t" + 
							expertcnt);
			}
		}
	}
	
	/***
	 * 
	 */	
	
	/**
	 * helper routine to convert ave dur of mpperformance notes in long to a float array
	 */
	protected ArrayList<Float> calcBaselineofPerformance(MPOriginalPerformance perf)
	{
		ArrayList<Float> baseline = new ArrayList<Float>(perf.getNumNotes());
		for (int i = 0; i < perf.getNumNotes(); i++) {
			float f = (float)perf.getNoteDur(i);
			baseline.add(f);
		}
		// special case:  some original midi short by one !
		if (perf.getNumNotes() < stats.maxNoteCount) {
			baseline.add((float)0);
		}
		return baseline;
	}
		
	public void ReportSongSlope()
	{		
		Collections.sort(songList.songs, new MPPerformanceSlopeComparator());

		out.MPWriteLn("");
		out.MPWriteLn(MPPerformance.SlopeFormat());
		for (MPPerformance perf : songList.songs) {
			perf.reportSlope(out);
		}
	}

	public void ReportExpertPerformances(MPOutputStream output, MPPhrase phrase, MPSongStats stats)
	{
		int cnt = 0;
		output.MPWriteLn("experts");
		for (MPPerformance p : songList.songs) {
			if (phrase != null && p.PhraseMatch(phrase, stats)) {
				p.Report(output);
			}
			if (++cnt > 20) {
				break;
			}
		}
	}

	public void ReportPerformances(MPOutputStream output)
	{
		int cnt = 0;
		output.MPWriteLn("");
		for (MPPerformance p : songList.songs) {
			p.Report(output);
			if (++cnt > 100) {
				break;
			}
		}
	}
	
	public void ReportPerformances()
	{
		ReportPerformances(out);
	}

}

