import java.util.*;

/***
 * 
 * @author jeffsmith
 *
 * a cluster of performances, stored by paired indexes into a performance list
 */
public class MPSpearmanArray {

	protected ArrayList<MPSpearmanItem> spearmanArray;
	
	/***
	 * 
	 * Constructor for MPSpearmanArray class
	 * 
	 * as input, take the array of note durations for two performances, xDur and yDur
	 * 
	 * construct a table, spearmanArray, of four columns (MPSpearmanItem), and num-notes rows
	 * 
	 *   	long xDur:  the given duration of X
   	 *		long yDur:  the given duration of Y
	 * 		xRank:	computed rank of xDur
	 * 		yRank:  computed rank of yDur
	 * 
	 */
	public MPSpearmanArray(ArrayList<Long> xDur, ArrayList<Long> yDur)
	{
		int max_size = Math.max(xDur.size(), yDur.size());
		spearmanArray = new ArrayList<MPSpearmanItem>(max_size);
		
		int size = Math.min(xDur.size(), yDur.size());
		for (int i = 0; i < size; i++) {
			spearmanArray.add(new MPSpearmanItem(xDur.get(i), yDur.get(i)));
		}
		
		if (xDur.size() > size) {
			for (int i = size; i < xDur.size(); i++) {
				spearmanArray.add(new MPSpearmanItem(xDur.get(i), 0));
			}
		} else if (yDur.size() > size) {
			for (int i = size; i < yDur.size(); i++) {
				spearmanArray.add(new MPSpearmanItem(0, yDur.get(i)));
			}
			
		}
	}

	/***
	 * Sort all X performances by duration and compute a rank, assigned to each Spearman array item
	 */
	protected void calcXRank()
	{
		Collections.sort(spearmanArray, new MPSpearmanItemXComparator());
		
		for (int i = 0; i < spearmanArray.size(); i++) {
			spearmanArray.get(i).SetXRank((short)i);
		}
	}
	
	/***
	 * Sort all Y performances by duration and compute a rank, again assigned to each Spearman array item
	 */
	protected void calcYRank()
	{
		Collections.sort(spearmanArray, new MPSpearmanItemYComparator());
		
		for (int i = 0; i < spearmanArray.size(); i++) {
			spearmanArray.get(i).SetYRank((short)i);
		}
	}

	/***
	 * calc ranks for X and Y, filling in the 3rd and 4th columns in the spearman table
	 * 
	 * then walk the table and compute the square of the sums of the ranks
	 */
	protected double calcDSquaredSum()
	{
		calcXRank();
		calcYRank();
	
		double dSquaredSum = 0;
		for (MPSpearmanItem item : spearmanArray) {
			long d = item.yRank - item.xRank;
			dSquaredSum += d * d;
		}
		
		return dSquaredSum;
	}
	
	/****
	 *
	 * Calc the Spearman rank by computing dSquared across the table, then applying the Spearman formula
	 *
	 */
	
	public float CalcSpearmanRank()
	{
		if (spearmanArray == null) {
			System.out.println("MPSpearmanArray: CalcDSquared: empty list");
			return -1;
		}
		
		double dSquaredSum = calcDSquaredSum();
		double n = spearmanArray.size();
		n = n * (n * n - 1.0);
		double p = 1 - ((6 * dSquaredSum) / n);
		return (float)p;
	}
}
