import java.util.ArrayList;
import java.util.TreeSet;


public abstract class RIRegionItem implements Comparable<RIRegionItem> {
	
	class RIItem {
		protected int id;
		protected float x;
		protected float y;
		protected MPPerformance p;
		
		public RIItem(RIItem i)
		{
			p = i.p;
			id = i.id;
			x = i.x;
			y = i.y;
		}
		
		public RIItem(MPPerformance perf)
		{
			p = perf;
			id = p.getUserID();
			x = p.getLoc().n_long;
			y = p.getLoc().n_lat;
		}
	}
	
	public String region_name;
	protected ArrayList<RIItem> perfs;
	public TreeSet<Integer> unique_ids;

	protected int total_performs;
	protected int expert_count;
	protected int urban_count;
	protected double average_note_duration;
	
	RIInterRegionItem inter_region;
	RIIntraRegionItem intra_region;
		
	public RIRegionItem()
	{
		perfs = new ArrayList<RIItem>();
		unique_ids = new TreeSet<Integer>();
		inter_region = new RIInterRegionItem();
		intra_region = new RIIntraRegionItem();
	}

	public String GetRegionName()
	{
		return region_name;
	}
	
	public int compareTo(RIRegionItem arg0) 
	{
		return region_name.compareTo(arg0.region_name);
	}
	
	public int GetNumUniques()
	{
		int uniques = unique_ids.size();
		return uniques;
	}
	
	public ArrayList<Integer> GetUniques()
	{
		ArrayList<Integer> unique_list = new ArrayList<Integer>(unique_ids);
		return unique_list;
	}
	
	public float GetUrbanPercent()
	{
		if (total_performs > 0) {
			return urban_count / (float)total_performs;
		}
		return 0f;
	}
	
	public float GetExpertPercent()
	{
		if (total_performs > 0) {
			return expert_count / (float)total_performs;
		}
		return 0f;
	}
	
	public float GetAverageTempo()
	{
		if (total_performs > 0) {
			return (float)(average_note_duration / (float)total_performs);
		}
		return 0f;
	}
	
	public String GetFullName()
	{
		return region_name;
	}
	
	public String GetRegionInfo()
	{
		return "";
	}

	public abstract boolean Contains(MPGeoCoord2 loc);
	
	public abstract boolean Contains(MPGeoCoord2 loc, double scale);

	protected class RIIntraRegionItem {
		
		// the number of comparisons used to produced this item
		protected int region_size;
		protected long num_compares;
		protected double pearson;

	    public RIIntraRegionItem()
	    {
	    }
	}
	
	protected class RIInterRegionItem extends RIIntraRegionItem {
		
		protected RIRegionItem region;

	    public RIInterRegionItem()
	    {
	    	super();
	    }
	}
	
	public static String GetRegionKey(int region_type, MPGeoCoord2 coord)
	{
		String key = null;
		switch (region_type) {
		case MPConfig.ANALYZE_BY_COUNTRY:
			key = RICountryRegionItem.GetKey(coord);
			break;
		case MPConfig.ANALYZE_BY_GEO:
			key = RIGeoRegionItem.GetKey(coord);
			break;
		case MPConfig.ANALYZE_BY_STATE:
			if (coord.isState()) {
				key = RIStateRegionItem.GetKey(coord);
			}
			break;
		case MPConfig.ANALYZE_BY_CITY:
			if (coord.isCity()) {
				key = RICityRegionItem.GetKey(coord);
			}
			break;
		case MPConfig.ANALYZE_BY_CONTINENT:
			key = RIContinentRegionItem.GetKey(coord);
			break;
		case MPConfig.ANALYZE_BY_LATITUDE:
			key = RILatitudeRegionItem.GetKey(coord);
			break;
		default:
			System.err.println("unknown region type");
			break;
		}
		return key;
	}
	
	public static RIRegionItem AllocRegionItem(int region_type, MPGeoCoord2 coord)
	{
		RIRegionItem item = null;
		switch (region_type) {
		case MPConfig.ANALYZE_BY_COUNTRY:
			item = new RICountryRegionItem(coord);
			break;
		case MPConfig.ANALYZE_BY_GEO:
			item = new RIGeoRegionItem(coord);
			break;
		case MPConfig.ANALYZE_BY_STATE:
			if (coord.isState()) {
				item = new RIStateRegionItem(coord);
			}
			break;
		case MPConfig.ANALYZE_BY_CITY:
			if (coord.isCity()) {
				item = new RICityRegionItem(coord);
			}
			break;
		case MPConfig.ANALYZE_BY_CONTINENT:
			item = new RIContinentRegionItem(coord);
			break;
		case MPConfig.ANALYZE_BY_LATITUDE:
			item = new RILatitudeRegionItem(coord);
			break;
		default:
			System.err.println("unknown region type");
			break;
		}
		if (item != null) {
			String regionname = item.GetRegionName();
			if (regionname == null || regionname.isEmpty()) {
				return null;
			}
		}
		return item;
	}
	
	public void AddInterRegion(RIRegionItem reg, int size, long compares, double r)
	{
		inter_region.pearson += compares * r;
		inter_region.num_compares += compares;
		inter_region.region_size += size; 

		// RIInterRegionItem iri = new RIInterRegionItem(reg, size, compares, r);
		// inter_regions.add(iri);
	}
	
	public void AddIntraRegion(int size, long compares, double r)
	{
		intra_region.pearson += compares * r;
		intra_region.num_compares += compares;
		intra_region.region_size += size; 

		// RIIntraRegionItem iri = new RIIntraRegionItem(size, compares, r);
		// intra_regions.add(iri);
	}

	public RIRegionData AggregateIntraRegions()
	{
		double pearson = 0;
		if (intra_region.num_compares > 0) {
			pearson = intra_region.pearson / (double)intra_region.num_compares;
		}
		RIRegionData rd = new RIRegionData(pearson, intra_region.num_compares, intra_region.region_size);
		return rd;
	}
	
	public RIRegionData AggregateInterRegions()
	{
		double pearson = 0;
		if (inter_region.num_compares > 0) {
			pearson = inter_region.pearson / (double)inter_region.num_compares;
		}
		RIRegionData rd = new RIRegionData(pearson, inter_region.num_compares, inter_region.region_size);
		return rd;
	}
	
	public void AddPerformance(RIItem i)
	{
		unique_ids.add(i.id);
		perfs.add(i);
	}
	
	public void AddPerformance(MPPerformance p)
	{
		unique_ids.add(p.getUserID());
		perfs.add(new RIItem(p));
	}
	
	public void ResetPerformances()
	{
		unique_ids.clear();
		perfs.clear();

		perfs = new ArrayList<RIItem>();
		unique_ids = new TreeSet<Integer>();
	}
	
	public void AddRegionUserData(MPSongList songlist, MPPhrase phrase, MPSongStats stats)
	{
		// move this to AddPerformance logic
		for (MPPerformance p : songlist.songs) {
			total_performs++;
			if (p.getLoc().is_city) {
				urban_count++;
			}
			if (phrase != null && p.PhraseMatch(phrase, stats)) {
				++expert_count;
			}
			
			// how fast relative to average tempo?
			double d = (double)p.noteAverageDuration / (double)stats.averageNoteLength;
			
			// aggregate this; will divide by total performers later to get average tempo in this region
			average_note_duration += d;
		}
		
	}
	
	/*
	 * find all songs in region which share songid
	 */
	
	public MPSongList getSongsInRegion(String songid)
	{
		// while constructing list, prune any songs that were pruned outside of region processing
		MPSongList list = new MPSongList();
		int i = 0;
		while (i < perfs.size()) {
			MPPerformance p = perfs.get(i).p;
			if (p.is_pruned) {
				perfs.remove(i);
			} else {
				if (p.GetPerformanceSongID().equals(songid)) {
					if (p.noteDur != null) {
						list.Add(p);
					}
				}
				++i;
			}
		}
		return list;
	}
	
	public void AddIntraRegionItems(RIRegionData data)
	{
		AddIntraRegion(data.size, data.compares, data.pearson);
	}

	public void AddInterRegionItems(RIRegionItem reg, RIRegionData data)
	{
		AddInterRegion(reg, data.size, data.compares, data.pearson);
	}

	public boolean isSuperSet(RIRegionItem item)
	{
		return perfs.containsAll(item.perfs);
	}
	
	public boolean Subdivide(ArrayList<RIRegionItem> newitems, int min_uniques, int max_uniques) 
	{
		// by default, a region doesn't know how to subdivide itself
		return false;
	}
}

