import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.TreeMap;

public class RIRegionItemSet {
	TreeMap<String, RIRegionItem> tree;
	ArrayList<RIRegionItem> list;
	
	public RIRegionItemSet() 
	{
		tree = new TreeMap<String, RIRegionItem>();
	}
	
	public int GetSize()
	{
		return tree.size();
	}
	
	public void Put(RIRegionItem ri)
	{
		tree.put(ri.GetRegionName(), ri);
	}
	
	public RIRegionItem Get(String key) 
	{
		return tree.get(key);
	}
	
	public RIRegionItem Get(RIRegionItem ri) 
	{
		return Get(ri.GetRegionName());
	}
	
	public Collection<RIRegionItem> GetCollection()
	{
		return tree.values();
	}

	public ArrayList<RIRegionItem> GetList()
	{
		if (list == null) {
			list = new ArrayList<RIRegionItem>(tree.values());
		}
		return list;
	}
	
	/*
	 * 
	 * we need all regions to be in the target range of min/max uniques
	 * 
	 * for any region > max, subdivide region until we achieve < max uniques
	 * 
	 */
	public void AdjustRegionSizes(int min_uniques, int max_uniques)
	{
		ArrayList<RIRegionItem> adds = new ArrayList<RIRegionItem>();
		ArrayList<RIRegionItem> deletes = new ArrayList<RIRegionItem>();
		Collection<RIRegionItem> c = tree.values();
		Iterator<RIRegionItem> i = c.iterator();
		while (i.hasNext()) {
			RIRegionItem ri = i.next();
			if (ri.GetNumUniques() < min_uniques) {
			} else {
				if (ri.GetNumUniques() > max_uniques) {
					
					// if we subdivide, we've divided this item into several new items,
					// which in turn replace this item
					if (ri.Subdivide(adds, min_uniques, max_uniques)) {
						deletes.add(ri);
					}
				}
			}
		}		
		
		for (RIRegionItem ri : adds) {
			Put(ri);
		}
		
		for (RIRegionItem ri : deletes) {
			tree.remove(ri.GetRegionName());
		}
		deletes.clear();
	}
	
	public void AllocateRegions(int region_type, ArrayList<MPSong> songs)
	{
		for (MPSong thesong : songs) {
			for (MPPerformance perf : thesong.songList.songs) {
				String key = RIRegionItem.GetRegionKey(region_type, perf.perfLoc);
				if (key != null) {

					// does this region item exist, yet, in the tree?
					RIRegionItem item = Get(key);

					// if not, create an item and add to tree
					if (item == null) {
						// System.out.println("new region " + key);
						item = RIRegionItem.AllocRegionItem(region_type, perf.perfLoc);
						Put(item);
					}
				}
			}
		}
	}

	public void Populate(ArrayList<MPSong> songs, int max_uniques)
	{
		// clear any existing performances
		Collection<RIRegionItem> c = GetCollection();
		Iterator<RIRegionItem> i = c.iterator();
		while (i.hasNext()) {
			RIRegionItem ri = i.next();
			ri.ResetPerformances();
		}
		
		// populate the tree with performance data from each song
		// start with smallest scale, and then expand (fill region from inside out)
		double scale = 1.0;
		// !!! * 2
		double target_scale = MPAnalyzer.GetConfig().geo_region_scale * 2;
		double increment = (target_scale - scale) / 3;
		if (increment <= 0) {
			increment = 1;
		}
		while (scale <= target_scale) {
			for (MPSong thesong : songs) {
				populateRegionPerformances(thesong, max_uniques, scale);
			}
			scale += increment;
		}
		
	}

	protected void populateRegionPerformances(MPSong song, int max_uniques, double scale)
	{
		Collection<RIRegionItem> c = GetCollection();
		Iterator<RIRegionItem> i = c.iterator();
		while (i.hasNext()) {
			RIRegionItem ri = i.next();
			for (MPPerformance p : song.songList.songs) {
				if (ri.GetNumUniques() < max_uniques) {
					if (ri.Contains(p.perfLoc, scale)) {
						ri.AddPerformance(p);
					}
				}
			}
		}		
	}
	
	protected void PruneRedundantRegions()
	{
		ArrayList<RIRegionItem> deletes = new ArrayList<RIRegionItem>();
		ArrayList<RIRegionItem> tmp_list = GetList();
		for (int i = 0; i < tmp_list.size(); i++) {
			RIRegionItem i_r = tmp_list.get(i);
			for (int j = i + i; j < tmp_list.size(); j++) {
				RIRegionItem j_r = tmp_list.get(j);
				if (i_r.isSuperSet(j_r)) {
					deletes.add(j_r);
				}
			}
		}

		for (RIRegionItem ri : deletes) {
			tree.remove(ri.GetRegionName());
		}
		deletes.clear();

		list.clear();
		list = null;
	}
	
	private ArrayList<RIRegionItem> getMinimumRegionsContainingCoord(MPGeoCoord2 coord)
	{
		// find the fewest and smallest region(s) which contain this coordinate
		ArrayList<RIRegionItem> region_item_list = GetList();	
		ArrayList<RIRegionItem> compare_regions = new ArrayList<RIRegionItem>();	
		double radius = 0.2;
		do {
			for (RIRegionItem ri : region_item_list) {
				// no scaling so we can find best match
				
				// take out all urban from comparisons.
				if (ri.Contains(coord, radius)) {
					compare_regions.add(ri);
				}
			}
			radius += .1;
		} while (compare_regions.size() < 1 && radius <= 1.0);
		
		// not found?
		if (compare_regions.size() < 1) {
			return null;
		}
				
		return compare_regions;
	}
	
	public void CorrelateToRegion(MPGeoCoord2 coord, ArrayList<MPSong> songs)
	{
		ArrayList<RIRegionItem> ri_list = GetList();
		ArrayList<RIRegionItem> coord_list = getMinimumRegionsContainingCoord(coord);
		if (coord_list == null) {
			System.out.println("CorrelateToRegion:  null regions");
			return;
		}

		for (MPSong thesong : songs) {
			System.out.println("loading " + thesong.songIO.songFileName + " " + thesong.songList.getSize());
			thesong.LoadSong();
			System.out.println("analyzing " + thesong.songIO.songFileName + " " + thesong.songList.getSize());
			thesong.CorrelateToRegion(coord_list, ri_list);
			thesong.Close();
		}
	}

	public void CorrelateByRegion(ArrayList<MPSong> songs)
	{
		ArrayList<RIRegionItem> ri_list = GetList();
		for (MPSong thesong : songs) {
			System.out.println("loading " + thesong.songIO.songFileName + " " + thesong.songList.getSize());
			thesong.LoadSong();
			System.out.println("analyzing " + thesong.songIO.songFileName + " " + thesong.songList.getSize());
			thesong.CorrelateByRegion(ri_list);
			thesong.Close();
		}
	}
	
	private class RIRegionReportItem implements Comparable<RIRegionReportItem> {
		String region_name;
		String region_info;
		RIRegionData inter_data;
		RIRegionData intra_data;
		int uniques;
		float urban;
		float expert;
		float tempo;
		ArrayList<Integer> unique_ids;
		
		public RIRegionReportItem(RIRegionItem ri)
		{
			region_name = ri.GetFullName();
			region_info = ri.GetRegionInfo();
			inter_data = ri.AggregateInterRegions();
			intra_data = ri.AggregateIntraRegions();
			uniques = ri.GetNumUniques();
			urban = ri.GetUrbanPercent();
			expert = ri.GetExpertPercent();
			tempo = ri.GetAverageTempo();
			unique_ids = ri.GetUniques();
		}
		
		// reverse order (highest to lowest)
		public int compareTo(RIRegionReportItem ri)
		{
			return (int)((ri.inter_data.pearson - inter_data.pearson) * 10000);
		}
		
		public String Report()
		{
			String s = region_name + "\t";
			s += region_info + "\t";
			s += inter_data.pearson + "\t";
			s += intra_data.pearson + "\t";
			s += inter_data.size + "\t";
			s += intra_data.size + "\t";
			s += uniques + "\t";
			s += urban + "\t";
			s += expert + "\t";
			s += tempo + "\t";
			return s;
		}
		
		public void ReportUniques(MPOutputStream out) 
		{
			for (Integer i : unique_ids) {
				out.MPWriteLn(i);
			}
		}
	}

	public void Report(MPOutputStream out, int min_uniques)
	{
		// iterate region tree.  aggregate inter and intra region values.  store in report array
		System.out.println("convert tree ");
		long miss = 0;
		ArrayList<RIRegionReportItem> report = new ArrayList<RIRegionReportItem>();
		GetList();
		System.out.println("iterate");
		for (RIRegionItem ri : list) {
			if (ri.GetNumUniques() >= min_uniques) {
				RIRegionReportItem item = new RIRegionReportItem(ri);
				report.add(item);
			} else {
				++miss;
			}
		}
		
		System.out.println("sort + miss " + miss);
		Collections.sort(report);

		out.MPWrite("region\tradius\t");
		out.MPWrite("inter pearson\tintra pearson\tinter size\tintra size\tuniques\t");
		out.MPWrite("urban%\texpert%\ttempo\t\n");
		out.MPWrite(MPAnalyzer.GetConfig().performance_max + "\t" + 
				MPAnalyzer.GetConfig().default_region_radius + "\t" + 
				MPAnalyzer.GetConfig().min_uniques + "\t" + 
				MPAnalyzer.GetConfig().max_uniques + "\t" + 
				MPAnalyzer.GetConfig().geo_region_size + " sqkm\t" + 
				MPAnalyzer.GetConfig().ANALYZE_SEGMENT_MAX + "\t" + 
				MPAnalyzer.GetConfig().compare_to_region + "\t" +
				MPAnalyzer.GetConfig().compare_to_region_str + "\n");	

		System.out.println("output");
		for (RIRegionReportItem rpi : report) {
			out.MPWriteLn(rpi.Report());
			// rpi.ReportUniques(out);
		}
	}
	
	public void OutputTenRegions(MPOutputStream out) 
	{
		int count = 0;
		Collection<RIRegionItem> c = tree.values();
		Iterator<RIRegionItem> i = c.iterator();
		while (i.hasNext() && count < 10) {
			RIRegionItem ri = i.next();
			if (ri.GetNumUniques() >= 40) {
				out.MPWrite(ri.GetRegionName() + "\t" + ri.inter_region.pearson + "\t\t" );
				for (int j = 0; j < 40; j++) {
					RIRegionItem.RIItem rii = ri.perfs.get(j);
					out.MPWrite(rii.p.noteAverageDuration + "\t");
				}
				out.MPWriteLn("");
				count++;
			}
		}
		
	}
}
